import numpy as np
from itertools import product

# from metrics.interface import omero as interface
from datetime import datetime
import logging

# import inspect
# from os import path
# import importlib
# importlib.resources.contents('metrics.samples')
# importlib.import_module('.argolight', package='metrics.samples')

# import dataset analysis
# from microscopemetrics.samples.dataset import DatasetConfigurator

# import devices
from microscopemetrics.devices import devices

# import samples
# from microscopemetrics.samples.argolight import ArgolightConfigurator
# from microscopemetrics.samples.psf_beads import PSFBeadsConfigurator

# SAMPLE_CONFIGURATORS = [ArgolightConfigurator,
#                         PSFBeadsConfigurator]
# noinspection PyUnresolvedReferences
# SAMPLE_HANDLERS = [c.SAMPLE_CLASS for c in SAMPLE_CONFIGURATORS]
# SAMPLE_SECTIONS = [c.CONFIG_SECTION for c in SAMPLE_CONFIGURATORS]
# SAMPLE_ANALYSES = [c.ANALYSES for c in SAMPLE_CONFIGURATORS]

# Creating logging services
module_logger = logging.getLogger("metrics.analysis")

# Namespace constants
NAMESPACE_PREFIX = "metrics"
NAMESPACE_ANALYZED = "analyzed"
NAMESPACE_VALIDATED = "validated"
# TODO: Add a special case editable


def get_image_data(image, device):
    image_name = image.getName()
    image_id = image.getId()
    raw_img = interface.get_intensities(image)
    # Images from Interface come in zctyx dimensions and locally as zcxy.
    # The easiest for the moment is to remove t
    if raw_img.shape[2] == 1:
        raw_img = np.squeeze(raw_img, 2)  # TODO: Fix this time dimension.
    else:
        raise Exception(
            "Image has a time dimension. Time is not yet implemented for this analysis"
        )
    pixel_size = interface.get_pixel_size(image)
    pixel_size_units = interface.get_pixel_size_units(image)

    device_settings = device.get_all_settings(image=image)

    image_settings = {
        "image_data": raw_img,
        "image_name": image_name,
        "image_id": image_id,
        "pixel_size": pixel_size,
        "pixel_size_units": pixel_size_units,
    }
    image_settings.update(device_settings)

    return image_settings


def save_data_table(
    conn, table_name, col_names, col_descriptions, col_data, interface_obj, namespace
):

    table_ann = interface.create_annotation_table(
        connection=conn,
        table_name=table_name,
        column_names=col_names,
        column_descriptions=col_descriptions,
        values=col_data,
        namespace=namespace,
    )

    interface_obj.linkAnnotation(table_ann)


def save_data_key_values(conn, key_values, interface_obj, namespace, editable=False):
    map_ann = interface.create_annotation_map(
        connection=conn, annotation=key_values, namespace=namespace
    )
    interface_obj.linkAnnotation(map_ann)


def create_roi(conn, shapes, image, name, description):
    type_to_func = {
        "point": interface.create_shape_point,
        "line": interface.create_shape_line,
        "rectangle": interface.create_shape_rectangle,
        "ellipse": interface.create_shape_ellipse,
        "polygon": interface.create_shape_polygon,
        "mask": interface.create_shape_mask,
    }

    new_shapes = [type_to_func[shape["type"]](**shape["args"]) for shape in shapes]
    interface.create_roi(
        connection=conn,
        image=image,
        shapes=new_shapes,
        name=name,
        description=description,
    )


def create_image(
    conn,
    image_intensities,
    image_name,
    description,
    dataset,
    source_image_id=None,
    metrics_generated_tag_id=None,
):

    zct_list = list(
        product(
            range(image_intensities.shape[0]),
            range(image_intensities.shape[1]),
            range(image_intensities.shape[2]),
        )
    )
    zct_generator = (image_intensities[z, c, t, :, :] for z, c, t in zct_list)

    new_image = conn.createImageFromNumpySeq(
        zctPlanes=zct_generator,
        imageName=image_name,
        sizeZ=image_intensities.shape[0],
        sizeC=image_intensities.shape[1],
        sizeT=image_intensities.shape[2],
        description=description,
        dataset=dataset,
        sourceImageId=source_image_id,
    )

    # TODO: see with ome why description is not applied
    img_wrapper = conn.getObject("Image", new_image.getId())
    img_wrapper.setDescription(description)
    img_wrapper.save()

    if metrics_generated_tag_id is not None:
        tag = conn.getObject("Annotation", metrics_generated_tag_id)
        if tag is None:
            module_logger.warning(
                "Metrics tag is not found. New images will not be tagged. Verify metrics tag existence and id."
            )
        else:
            new_image.linkAnnotation(tag)

    return new_image


def analyze_dataset(connection, script_params, dataset, analysis_config, device_config):
    # TODO: must note in mapann the analyses that were done

    # TODO: do something to automate selection of microscope type
    device = devices.WideFieldMicroscope(device_config)

    module_logger.info(f"Analyzing data from Dataset: {dataset.getId()}")
    module_logger.info(f"Date and time: {datetime.now()}")

    # This dictionary keeps track if the hard or soft limits are passed or not for the whole dataset
    dataset_limits_passed = {
        "uhl_passed": True,
        "lhl_passed": True,
        "usl_passed": True,
        "lsl_passed": True,
        "limits": list(),
        "sources": list(),
    }

    for section, analyses, handler in zip(
        SAMPLE_SECTIONS, SAMPLE_ANALYSES, SAMPLE_HANDLERS
    ):
        if analysis_config.has_section(section):
            module_logger.info(f"Running analysis on {section.capitalize()} sample(s)")
            section_conf = analysis_config[section]
            handler_instance = handler(config=section_conf)
            for analysis in analyses:
                if section_conf.getboolean(f"analyze_{analysis}"):
                    namespace = (
                        f"{NAMESPACE_PREFIX}/"
                        f"{NAMESPACE_ANALYZED}/"
                        f"{handler.get_module()}/"
                        f"{analysis}/"
                        f'{analysis_config["MAIN"]["metrics_version"]}'
                    )
                    images = interface.get_tagged_images_in_dataset(
                        dataset, section_conf.getint(f"{analysis}_image_tag_id")
                    )
                    for image in images:
                        image_data = get_image_data(image=image, device=device)
                        (
                            out_images,
                            out_rois,
                            out_tags,
                            out_dicts,
                            out_tables,
                            image_limits_passed,
                        ) = handler_instance.analyze_image(
                            image=image_data, analyses=analysis, config=section_conf
                        )

                        for out_image in out_images:
                            create_image(
                                conn=connection,
                                image_intensities=out_image["image_data"],
                                image_name=out_image["image_name"],
                                description=f'Source Image Id:{image.getId()}\n{out_image["image_desc"]}',
                                dataset=dataset,
                                source_image_id=image.getId(),
                                metrics_generated_tag_id=analysis_config["MAIN"].getint(
                                    "metrics_generated_tag_id"
                                ),
                            )

                        for out_roi in out_rois:
                            create_roi(
                                conn=connection,
                                shapes=out_roi["shapes"],
                                image=image,
                                name=out_roi["name"],
                                description=out_roi["desc"],
                            )

                        for out_tag in out_tags:
                            pass  # TODO implement interface to save tags

                        for out_table_name, out_table in out_tables.items():
                            save_data_table(
                                conn=connection,
                                table_name=out_table_name,
                                col_names=[p["name"] for p in out_table],
                                col_descriptions=[p["desc"] for p in out_table],
                                col_data=[p["data"] for p in out_table],
                                interface_obj=image,
                                namespace=namespace,
                            )

                        for out_dict in out_dicts:
                            labeled_out_dict = {
                                "analysis_date_time": f"{datetime.now()}",
                                "sample_type": f"{handler.get_module()}",
                                "analysis_type": f"{analysis}",
                            }
                            labeled_out_dict.update(out_dict)

                            save_data_key_values(
                                conn=connection,
                                key_values=labeled_out_dict,
                                interface_obj=image,
                                namespace=namespace,
                            )

                        for ilp in image_limits_passed:
                            for k, v in ilp.items():
                                if v is False:
                                    dataset_limits_passed[k] = False
                                elif type(v) is list:
                                    dataset_limits_passed[k].extend(v)

    dataset_section = DatasetConfigurator.CONFIG_SECTION
    dataset_analyses = DatasetConfigurator.ANALYSES
    dataset_handler = DatasetConfigurator.SAMPLE_CLASS

    if analysis_config.has_section(dataset_section):
        module_logger.info(f"Running analysis on dataset")
        ds_conf = analysis_config[dataset_section]
        ds_handler_instance = dataset_handler(config=ds_conf)
        for dataset_analysis in dataset_analyses:
            if ds_conf.getboolean(f"analyze_{dataset_analysis}"):
                namespace = (
                    f"{NAMESPACE_PREFIX}/"
                    f"{NAMESPACE_ANALYZED}/"
                    f"{dataset_handler.get_module()}/"
                    f"{dataset_analysis}/"
                    f'{analysis_config["MAIN"]["metrics_version"]}'
                )
                (
                    out_images,
                    out_tags,
                    out_dicts,
                    out_editables,
                    out_tables,
                    image_limits_passed,
                ) = ds_handler_instance.analyze_dataset(
                    dataset=dataset, analyses=dataset_analysis, config=ds_conf
                )
                for out_image in out_images:
                    create_image(
                        conn=connection,
                        image_intensities=out_image["image_data"],
                        image_name=out_image["image_name"],
                        description=out_image["image_desc"],
                        dataset=dataset,
                        metrics_generated_tag_id=analysis_config["MAIN"].getint(
                            "metrics_generated_tag_id"
                        ),
                    )

                for out_tag in out_tags:
                    pass  # TODO implement interface to save tags

                for out_table_name, out_table in out_tables.items():
                    save_data_table(
                        conn=connection,
                        table_name=out_table_name,
                        col_names=[p["name"] for p in out_table],
                        col_descriptions=[p["desc"] for p in out_table],
                        col_data=[p["data"] for p in out_table],
                        interface_obj=dataset,
                        namespace=namespace,
                    )

                for out_dict, out_editable in zip(out_dicts, out_editables):
                    if out_editable:
                        tmp_namespace = None
                    else:
                        tmp_namespace = namespace
                    save_data_key_values(
                        conn=connection,
                        key_values=out_dict,
                        interface_obj=dataset,
                        namespace=tmp_namespace,
                    )

                for ilp in image_limits_passed:
                    for k, v in ilp.items():
                        if v is False:
                            dataset_limits_passed[k] = False
                        elif type(v) is list:
                            dataset_limits_passed[k].extend(v)

    # Save final dataset limits passed tests and corresponding tags
    namespace = (
        f"{NAMESPACE_PREFIX}/"
        f"{NAMESPACE_ANALYZED}/"
        "dataset/"
        "limits_verification/"
        f'{analysis_config["MAIN"]["metrics_version"]}'
    )

    # dataset_limits_passed['limits'] = list(dict.fromkeys(dataset_limits_passed['limits']))  # Remove duplicates

    if (
        dataset_limits_passed["uhl_passed"] and dataset_limits_passed["lhl_passed"]
    ):  # Hard limits passed
        interface.link_annotation_tag(
            connection,
            dataset,
            analysis_config["MAIN"].getint("passed_hard_limits_tag_id"),
        )
    else:
        interface.link_annotation_tag(
            connection,
            dataset,
            analysis_config["MAIN"].getint("not_passed_hard_limits_tag_id"),
        )

    if (
        dataset_limits_passed["usl_passed"] and dataset_limits_passed["lsl_passed"]
    ):  # Soft limits passed
        interface.link_annotation_tag(
            connection,
            dataset,
            analysis_config["MAIN"].getint("passed_soft_limits_tag_id"),
        )
    else:
        interface.link_annotation_tag(
            connection,
            dataset,
            analysis_config["MAIN"].getint("not_passed_soft_limits_tag_id"),
        )

    save_data_key_values(
        conn=connection,
        key_values=dataset_limits_passed,
        interface_obj=dataset,
        namespace=namespace,
    )

    try:
        if "Comment" in script_params.keys() and script_params["Comment"] != "":
            module_logger.info("Adding comment to Dataset.")
            namespace = (
                f"{NAMESPACE_PREFIX}/"
                f"{NAMESPACE_ANALYZED}/"
                "comment/"
                "comment/"
                f'{analysis_config["MAIN"]["config_version"]}'
            )

            comment_annotation = interface.create_annotation_comment(
                connection=connection,
                comment_string=script_params["Comment"],
                namespace=namespace,
            )
            interface.link_annotation(dataset, comment_annotation)
    except KeyError:
        module_logger.info("No comments added")

    module_logger.info(f"Sample finished for dataset: {dataset.getId()}")



import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np
from scipy.interpolate import griddata


def plot_distances_maps(distances, x_dim, y_dim):
    """[((ch_A, ch_B), [[(s_x, s_y, s_z), dst, t_index],...]),...]"""
    nb_of_channels = 4
    fig, axes = plt.subplots(
        ncols=nb_of_channels, nrows=nb_of_channels, squeeze=False, figsize=(12, 12)
    )

    for p in distances:
        positions_map = np.asarray([(x, y) for (z, x, y) in p["coord_of_A"]])
        distances_map = np.asarray(p["dist_3d"])

        grid_x, grid_y = np.mgrid[0:x_dim:1, 0:y_dim:1]
        interpolated = griddata(
            positions_map, distances_map, (grid_x, grid_y), method="cubic"
        )

        ax = axes.ravel()
        ax[(p["channels"][0] * 4) + p["channels"][1]].imshow(
            np.flipud(interpolated),
            extent=(0, x_dim, y_dim, 0),
            origin="lower",
            cmap=cm.hot,
            # vmin=np.amin(raw_stack[0, :, c, :, :]),
            # vmax=np.amax(raw_stack[0, :, c, :, :])
        )

    plt.show()


def plot_homogeneity_map(raw_stack, spots_properties, spots_positions, labels_stack):

    nb_of_channels = raw_stack.shape[1]
    x_dim = raw_stack.shape[-2]
    y_dim = raw_stack.shape[-1]

    fig, axes = plt.subplots(
        ncols=nb_of_channels, nrows=3, squeeze=False, figsize=(12, 6)
    )

    for c in range(nb_of_channels):
        weighted_centroid = np.array(
            [x["weighted_centroid"][0] for x in spots_properties[c]]
        )
        areas = np.array([x["area"] for x in spots_properties[c]])
        max_intensity = np.array([x["max_intensity"] for x in spots_properties[c]])
        grid_x, grid_y = np.mgrid[0:x_dim, 0:y_dim]
        try:
            interpolated = griddata(
                spots_positions[c][:, 1:],
                max_intensity,
                (grid_x, grid_y),
                method="linear",
            )
        except Exception as e:
            # TODO: Log a warning
            interpolated = np.zeros((256, 256))

        ax = axes.ravel()
        ax[c] = plt.subplot(3, 4, c + 1)

        ax[c].imshow(raw_stack[..., c, :, :].max(0), cmap="gray")
        ax[c].set_title("raw_channel_" + str(c))

        ax[c + nb_of_channels].imshow(labels_stack[..., c, :, :].max(0))
        ax[c + nb_of_channels].set_title("segmented_channel_" + str(c))

        ax[c + 2 * nb_of_channels].imshow(
            np.flipud(interpolated),
            extent=(0, x_dim, y_dim, 0),
            origin="lower",
            cmap=cm.hot,
            vmin=np.amin(raw_stack[..., c, :, :]),
            vmax=np.amax(raw_stack[..., c, :, :]),
        )
        ax[c + 2 * nb_of_channels].plot(
            spots_positions[c][:, 2], spots_positions[c][:, 1], "k.", ms=2
        )
        # ax[c + 2 * nb_of_channels].clim(np.amin(raw_img[:, c, :, :]), np.amax(raw_img[:, c, :, :]))
        ax[c + 2 * nb_of_channels].set_title("Max_intensity_channel_" + str(c))

    plt.show()


def plot_peaks(profiles, peaks, properties, resolutions, res_indexes):
    fig, axes = plt.subplots(
        ncols=1, nrows=len(profiles), squeeze=False, figsize=(48, 24)
    )

    for c, profile in enumerate(profiles):

        ax = axes.ravel()

        ax[c].plot(profile)
        ax[c].plot(peaks[c], profile[peaks[c]], "x")
        ax[c].vlines(
            x=peaks[c],
            ymin=profile[peaks[c]] - properties[c]["prominences"],
            ymax=profile[peaks[c]],
            color="green",
        )
        for i in res_indexes[c]:
            ax[c].vlines(
                x=peaks[c][i],
                ymin=profile[peaks[c][i]] + 0.02,
                ymax=1,
                color="red",
                linestyles="dashed",
            )
            ax[c].vlines(
                x=peaks[c][i + 1],
                ymin=profile[peaks[c][i + 1]] + 0.02,
                ymax=1,
                color="red",
                linestyles="dashed",
            )
            ax[c].annotate(
                s=f"{resolutions[c]:.3f}",
                xy=(
                    ((peaks[c][i] + peaks[c][i + 1]) / 2),
                    profile[peaks[c][i + 1]] + 0.1,
                ),
                # ((vline_value - x_bounds[0]) / (x_bounds[1] - x_bounds[0])), 1.01),
                # xycoords='axes fraction', verticalalignment='right',
                horizontalalignment="center",
                fontsize="xx-large",
                fontweight="bold",
                # rotation=270
            )

    # plt.legend()
    plt.show()



from enum import Enum, EnumMeta
from configparser import NoOptionError
from math import sin, asin, cos


# from microscopemetrics.interface import omero as interface

# Creating logging services
import logging

module_logger = logging.getLogger("metrics.devices.devices")


DTYPES = {
    "int": (tuple,),
    "float": (tuple,),
    "bool": (type(None),),
    "str": (int,),
    "tuple": (type(None),),
}


def _call_if_callable(f):
    """Call callables, or return value of non-callables."""
    return f() if callable(f) else f


class _Setting:
    """The code of this class was copied from https://github.com/MicronOxford/microscope"""

    # Settings classes should be private: devices should use a factory method
    # rather than instantiate settings directly; most already use add_setting
    # for this.
    # TODO: Implement a get units
    def __init__(
        self,
        name,
        dtype,
        get_from_db_func,
        get_from_conf_func,
        get_from_name_func,
        set_func,
        values=None,
    ):
        """Create a setting.
        :param name: the setting's name
        :param dtype: a data type from ('int', 'float', 'bool', 'enum', 'str')
        :param get_from_db_func: a function to get the value from the database interface
        :param get_from_conf_func: a function to get the value from the device configuration
        :param get_from_name_func: a function to get the value from the image name
        :param values: a description of allowed values dependent on dtype,
                       or function that returns a description
        Setters and getters accept or return:
            the setting value for int, float, bool and str;
            the setting index into a list, dict or Enum type for enum.
        """
        self.name = name
        if dtype not in DTYPES:
            raise Exception("Unsupported dtype.")
        elif not (isinstance(values, DTYPES[dtype]) or callable(values)):
            raise Exception(
                "Invalid values type for %s '%s':"
                " expected function or %s" % (dtype, name, DTYPES[dtype])
            )
        self.dtype = dtype
        self._get_from_db = get_from_db_func
        self._get_from_conf = get_from_conf_func
        self._get_from_name = get_from_name_func
        self._values = values

    def describe(self):
        """Returns a dictionary describing the setting"""
        return {"type": self.dtype, "values": self.values()}

    def get(self, **kwargs):
        """Get a setting"""
        if self._get_from_db is not None:
            value = self._get_from_db(**kwargs)
            if value is not None:
                return value
        if self._get_from_conf is not None:
            value = self._get_from_conf(**kwargs)
            if value is not None:
                return value
        if self._get_from_name is not None:
            value = self._get_from_name(**kwargs)
            if value is not None:
                return value
        module_logger.warning(f"Device setting could not be retrieved: {self.name}")

    def values(self):
        values = _call_if_callable(self._values)
        if values is not None:
            if self.dtype == "enum":
                if isinstance(values, dict):
                    return list(values.items())
                else:
                    # self._values is a list or tuple
                    return list(enumerate(values))
            elif self._values is not None:
                return values


class Device:
    """A superclass for all the microscope devices and eventually other type of devices."""

    def __init__(self, device_config):
        self.device_config = device_config
        self._settings = {}

    def add_setting(
        self,
        name,
        dtype,
        get_from_db_func,
        get_from_conf_func,
        get_from_name_func,
        set_func,
        values=None,
    ):
        """Add a setting definition.
        :param name: the setting's name
        :param dtype: a data type from ('int', 'float', 'bool', 'enum', 'str')
        :param get_from_db_func: a function to get the value from the interface to the database
        :param get_from_conf_func: a function to get the value from the configuration
        :param get_from_name_func: a function to get the value from the image name
        :param set_func: a function to set the value
        :param values: a description of allowed values dependent on dtype,
                       or function that returns a description
        A client needs some way of knowing a setting name and data type,
        retrieving the current value and, if settable, a way to retrieve
        allowable values, and set the value.
        """
        if dtype not in DTYPES:
            raise Exception("Unsupported dtype.")
        elif not (isinstance(values, DTYPES[dtype]) or callable(values)):
            raise Exception(
                "Invalid values type for %s '%s':"
                " expected function or %s" % (dtype, name, DTYPES[dtype])
            )
        else:
            self._settings[name] = _Setting(
                name,
                dtype,
                get_from_db_func,
                get_from_conf_func,
                get_from_name_func,
                set_func,
                values,
            )

    def get_setting(self, name, **kwargs):
        """Tries to get a specified setting from the following sources by order of preference:
        - Image from the database interface
        - The microscope analysis_config linked through the image name
        - Directly in the image name"""

        try:
            return self._settings[name].get(**kwargs)
        except Exception as err:
            module_logger.error("in get_setting(%s):", name, exc_info=err)
            raise

    def get_all_settings(self, **kwargs):
        """Return settings as a dict."""

        def catch(f):
            try:
                return f(**kwargs)
            except Exception as err:
                module_logger.error(f"getting {f.__self__.name}: {err}")
                return None

        return {k: catch(v.get) for k, v in self._settings.items()}

    def set_setting(self, name, value):
        """Set a setting."""
        try:
            self._settings[name].set(value)
        except Exception as err:
            module_logger.error("in set_setting(%s):", name, exc_info=err)
            raise

    def describe_setting(self, name):
        """Return ordered setting descriptions as a list of dicts."""
        return self._settings[name].describe()

    def describe_settings(self):
        """Return ordered setting descriptions as a list of dicts."""
        return [(k, v.describe()) for (k, v) in self._settings.items()]

    def settings_to_conf(self):
        pass

    def update_settings(self, incoming, init=False):
        """Update settings based on dict of settings and values."""
        if init:
            # Assume nothing about state: set everything.
            my_keys = set(self._settings.keys())
            their_keys = set(incoming.keys())
            update_keys = my_keys & their_keys
            if update_keys != my_keys:
                missing = ", ".join([k for k in my_keys - their_keys])
                msg = "update_settings init=True but missing keys: %s." % missing
                _logger.debug(msg)
                raise Exception(msg)
        else:
            # Only update changed values.
            my_keys = set(self._settings.keys())
            their_keys = set(incoming.keys())
            update_keys = set(
                key
                for key in my_keys & their_keys
                if self.get_setting(key) != incoming[key]
            )
        results = {}
        # Update values.
        for key in update_keys:
            if key not in my_keys or not self._settings[key].set:
                # Setting not recognised or no set function implemented
                results[key] = NotImplemented
                update_keys.remove(key)
                continue
            if _call_if_callable(self._settings[key].readonly):
                continue
            self._settings[key].set(incoming[key])
        # Read back values in second loop.
        for key in update_keys:
            results[key] = self._settings[key].get()
        return results


class Microscope(Device):
    """A superclass for the microscopes. Inherit this class when you create a new type of microscope."""

    def __init__(self, device_config):
        super().__init__(device_config)

    def _get_conf_objective_nr(self, image):
        img_name = image.getName()
        obj_nrs = [
            i
            for i, token in enumerate(self.device_config.getlist("OBJECTIVES", "names"))
            if token in img_name
        ]
        if len(obj_nrs) > 1:
            module_logger.error(
                "More than one reference to an objective lens was found in the image name. Only the first one will be considered."
            )
        elif len(obj_nrs) == 0:
            module_logger.info(
                "No references to any objective lens were found in the image name"
            )
            return None
        return obj_nrs[0]

    def _get_conf_objective_setting(self, option, image):
        obj_nr = self._get_conf_objective_nr(image)
        if obj_nr is None:
            return None
        try:
            values = eval(self.device_config.get("OBJECTIVES", option))
        except NoOptionError as e:
            module_logger.error(
                f"No parameters for {option} have been defined in the device configuration"
            )

        except NameError as e:
            module_logger.error(
                f"There was an error reading microscope objectives configuration option: {option}"
            )
            return None
        if values is None:
            module_logger.info(
                f"No information available for {option} in the microscope objectives configuration "
            )
            return None
        else:
            return values[obj_nr]

    def _get_conf_channel_nrs(self, image):
        img_name = image.getName()
        ch_nrs = list()
        token_positions = list()
        for ch, ch_token in enumerate(self.device_config.getlist("CHANNELS", "names")):
            token_pos = img_name.find(ch_token)
            if token_pos == -1:  # token not in name
                continue
            else:
                token_positions.append(token_pos)
                ch_nrs.append(ch)
        # Sort channels according to their position in the name
        ch_nrs = tuple(ch for _, ch in sorted(zip(token_positions, ch_nrs)))
        if len(ch_nrs) == 0:
            module_logger.info(
                "No references to any channel were found in the image name"
            )
            return None
        return ch_nrs

    def _get_conf_channel_settings(self, option, image):
        ch_nrs = self._get_conf_channel_nrs(image)
        if ch_nrs is None:
            return None
        try:
            values = eval(self.device_config.get("CHANNELS", option))
        except NameError as e:
            module_logger.error(
                f"There was an error reading microscope channel configuration option: {option}"
            )
            return None
        if values is None:
            module_logger.info(
                f"No information available for {option} in the microscope channels configuration "
            )
            return None
        else:
            return tuple(values[ch] for ch in ch_nrs)

    def _get_metadata_from_name(self, token_left, token_right, metadata_type, image):
        name = image.getName()
        start = name.find(token_left)
        if start == -1:
            return None
        name = name[start + len(token_left) :]
        end = name.find(token_right)
        value = metadata_type(name[:end])
        if metadata_type is None:
            return value
        else:
            try:
                value = metadata_type(value)
            except ValueError as e:
                module_logger.error(
                    f"Could not cast {value} into {metadata_type}. Please verify file naming for token {token_left}"
                )
                return None

        return value


class WideFieldMicroscope(Microscope):
    """A Widefield microscope"""

    def __init__(self, device_config):
        super().__init__(device_config)

        # Setting some objective lens settings
        self.add_setting(
            "objective_lens_refractive_index",
            "float",
            # get_from_db_func=interface._get_objective_lens_refractive_index,
            get_from_db_func=self._get_objective_lens_refractive_index,
            get_from_conf_func=self._get_conf_objective_lens_refr_index,
            get_from_name_func=self._get_name_objective_lens_refr_index,
            set_func=None,
            values=(1.0, 2.0),
        )
        self.add_setting(
            "objective_lens_na",
            "float",
            # get_from_db_func=interface._get_objective_lens_na,
            get_from_db_func=self._get_objective_lens_na,
            get_from_conf_func=self._get_conf_objective_lens_na,
            get_from_name_func=self._get_name_objective_lens_na,
            set_func=None,
            values=(0.0, 2.0),
        )
        self.add_setting(
            "objective_lens_nominal_magnification",
            "float",
            # get_from_db_func=interface._get_objective_lens_nominal_magnification,
            get_from_db_func=self._get_objective_lens_nominal_magnification,
            get_from_conf_func=self._get_conf_objective_lens_nominal_magnification,
            get_from_name_func=self._get_name_objective_lens_nominal_magnification,
            set_func=None,
            values=(1.0, 1000.0),
        )

        # Setting some channel settings
        self.add_setting(
            "excitation_wavelengths",
            "tuple",
            # get_from_db_func=interface._get_excitation_wavelengths,
            get_from_db_func=self._get_excitation_wavelengths,
            get_from_conf_func=self._get_conf_excitation_wavelengths,
            get_from_name_func=self._get_name_excitation_wavelengths,
            set_func=None,
            values=None,
        )
        self.add_setting(
            "emission_wavelengths",
            "tuple",
            # get_from_db_func=interface._get_emission_wavelengths,
            get_from_db_func=self._get_emission_wavelengths,
            get_from_conf_func=self._get_conf_emission_wavelengths,
            get_from_name_func=self._get_name_emission_wavelengths,
            set_func=None,
            values=None,
        )

    def get_all_settings(self, **kwargs):
        settings = super().get_all_settings(**kwargs)
        settings.update(self.get_theoretical_res(**kwargs))
        settings.update(self.get_nyquist(**kwargs))

        return settings

    def _get_conf_objective_lens_refr_index(self, image):
        return self._get_conf_objective_setting(
            "objective_lens_refractive_index", image
        )

    def _get_name_objective_lens_refr_index(self, image):
        return self._get_metadata_from_name("_RI=", "_", float, image)

    def _get_conf_objective_lens_na(self, image):
        return self._get_conf_objective_setting("objective_lens_na", image)

    def _get_name_objective_lens_na(self, image):
        return self._get_metadata_from_name("_NA=", "_", float, image)

    def _get_conf_objective_lens_nominal_magnification(self, image):
        return self._get_conf_objective_setting(
            "objective_lens_nominal_magnification", image
        )

    def _get_name_objective_lens_nominal_magnification(self, image):
        return self._get_metadata_from_name("_MAG=", "_", float, image)

    def _get_conf_excitation_wavelengths(self, image):
        return self._get_conf_channel_settings("excitation_wavelengths", image)

    def _get_name_excitation_wavelengths(self, image):
        return self._get_metadata_from_name("_EX=", "_", list, image)

    def _get_conf_emission_wavelengths(self, image):
        return self._get_conf_channel_settings("emission_wavelengths", image)

    def _get_name_emission_wavelengths(self, image):
        return self._get_metadata_from_name("_EM=", "_", list, image)

    # TODO: to move to interface
    def _get_objective_lens_refractive_index(self, image):
        objective_settings = image.getObjectiveSettings()
        if objective_settings is None:
            return None
        return objective_settings.getRefractiveIndex()

    def _get_objective_lens_na(self, image):
        objective_settings = image.getObjectiveSettings()
        if objective_settings is None:
            return None
        objective = objective_settings.getObjective()
        if objective is None:
            return None
        return objective.getLensNA()

    def _get_objective_lens_nominal_magnification(self, image):
        objective_settings = image.getObjectiveSettings()
        if objective_settings is None:
            return None
        objective = objective_settings.getObjective()
        if objective is None:
            return None
        return objective.getNominalMagnification()

    def _get_excitation_wavelengths(self, image):
        channels = image.getChannels()
        if channels is None:
            return None
        wavelengths = tuple(ch.getExcitationWave() for ch in channels)
        if not all(wavelengths):
            return None
        else:
            return wavelengths

    def _get_emission_wavelengths(self, image):
        channels = image.getChannels()
        if channels is None:
            return None
        wavelengths = tuple(ch.getEmissionWave() for ch in channels)
        if not all(wavelengths):
            return None
        else:
            return wavelengths

    def get_theoretical_res_fwhm(self, **kwargs):
        """Returns a dictionary containing the lateral and axial FWHM theoretical resolutions for every channel in image.

        :param image: an image object
        :return: a dictionary in the form:
                 {'resolution_theoretical_fwhm_lateral': list of lateral resolutions for every channel,
                  'resolution_theoretical_fwhm_axial': list of axial resolutions for every channel,
                  'resolution_theoretical_fwhm_units': string specifying units}
        """
        theoretical_res = {
            "resolution_theoretical_fwhm_lateral": [],
            "resolution_theoretical_fwhm_axial": [],
            "resolution_theoretical_fwhm_units": "NANOMETER",
        }
        na = self.get_setting("objective_lens_na", **kwargs)
        for em in self.get_setting("emission_wavelengths", **kwargs):
            try:
                theoretical_res["resolution_theoretical_fwhm_lateral"].append(
                    0.353 * em / na
                )
            except TypeError as e:
                module_logger.warning(
                    "FWHM theoretical resolution could not be calculated. Verify configuration files."
                )
                theoretical_res["resolution_theoretical_fwhm_lateral"].append(None)
            try:
                theoretical_res["resolution_theoretical_fwhm_axial"].append(
                    None
                )  # TODO: find formula for fwhm axial resolution
            except TypeError as e:
                module_logger.warning(
                    "FWHM theoretical resolution could not be calculated. Verify configuration files."
                )
                theoretical_res["resolution_theoretical_fwhm_axial"].append(None)

        return theoretical_res

    def get_theoretical_res_rayleigh(self, **kwargs):
        """Returns a dictionary containing the lateral and axial Rayleigh theoretical resolutions for every channel in image.

        :param image: an image object
        :return: a dictionary in the form:
                 {'resolution_theoretical_rayleigh_lateral': list of lateral resolutions for every channel,
                  'resolution_theoretical_rayleigh_axial': list of axial resolutions for every channel,
                  'resolution_theoretical_rayleigh_units': string specifying units}
        """
        theoretical_res = {
            "resolution_theoretical_rayleigh_lateral": [],
            "resolution_theoretical_rayleigh_axial": [],
            "resolution_theoretical_rayleigh_units": "NANOMETER",
        }
        na = self.get_setting("objective_lens_na", **kwargs)
        ri = self.get_setting("objective_lens_refractive_index", **kwargs)
        for em in self.get_setting("emission_wavelengths", **kwargs):
            try:
                theoretical_res["resolution_theoretical_rayleigh_lateral"].append(
                    0.61 * em / na
                )
            except TypeError as e:
                module_logger.warning(
                    "Rayleigh theoretical lateral resolution could not be calculated. Verify configuration files."
                )
                theoretical_res["resolution_theoretical_rayleigh_lateral"].append(None)
            try:
                theoretical_res["resolution_theoretical_rayleigh_axial"].append(
                    2 * em * ri / na**2
                )
            except TypeError as e:
                module_logger.warning(
                    "Rayleigh theoretical axial resolution could not be calculated. Verify configuration files."
                )
                theoretical_res["resolution_theoretical_rayleigh_axial"].append(None)

        return theoretical_res

    def get_theoretical_res(self, **kwargs):
        """Returns a dictionary containing the theoretical resolutions for every channel in image in all the
        implemented methods."""
        theoretical_res = dict()
        theoretical_res.update(self.get_theoretical_res_fwhm(**kwargs))
        theoretical_res.update(self.get_theoretical_res_rayleigh(**kwargs))

        return theoretical_res

    def get_nyquist(self, **kwargs):
        """Returns a dictionary containing the nyquist sampling criteria values for every channel in image."""
        nyquist_delta = {
            "nyquist_lateral": [],
            "nyquist_axial": [],
            "nyquist_units": "NANOMETER",
        }  # TODO: FIX units for nyquist and resolution. Get them from wavelength units.
        na = self.get_setting("objective_lens_na", **kwargs)
        ri = self.get_setting("objective_lens_refractive_index", **kwargs)
        for em in self.get_setting("emission_wavelengths", **kwargs):
            try:
                nyquist_delta["nyquist_lateral"].append(em / (4 * ri * (na / ri)))
            except TypeError as e:
                module_logger.warning(
                    "Lateral Nyquist criterion could not be calculated. Verify configuration files."
                )
                nyquist_delta["nyquist_lateral"].append(None)
            try:
                nyquist_delta["nyquist_axial"].append(
                    em / (2 * ri * (1 - cos(asin(na / ri))))
                )
            except TypeError as e:
                module_logger.warning(
                    "Axial Nyquist criterion could not be calculated. Verify configuration files."
                )
                nyquist_delta["nyquist_axial"].append(None)

        return nyquist_delta
        #
        # def calculate_theoretcal_resolution(microscope_type, na, refractive_index, emission_wave, excitation_wave=None):
        #     theoretical_res = {}
        #     if microscope_type is None:
        #         module_logger.warning(
        #             'Microscope type undefined to calculate theoretical resolution. Falling back into Wide-Field')
        #         theoretical_res = calculate_theoretcal_resolution('wf', na, refractive_index, emission_wave,
        #                                                           excitation_wave=excitation_wave)
        #     elif microscope_type.lower() in ['wf', 'wide-field', 'widefield']:
        #         theoretical_res['fwhm_lateral'] = .353 * emission_wave / na
        #         theoretical_res['rayleigh_lateral'] = .61 * emission_wave / na
        #         theoretical_res['rayleigh_axial'] = 2 * emission_wave * refractive_index / na ** 2
        #         theoretical_res['units'] = 'NANOMETER'
        #     elif microscope_type.lower() in ['confocal']:
        #         theoretical_res['fwhm_lateral'] = .353 * emission_wave / na
        #         theoretical_res['rayleigh_lateral'] = .4 * emission_wave / na
        #         theoretical_res['rayleigh_axial'] = 1.4 * emission_wave * refractive_index / na ** 2
        #         theoretical_res['units'] = 'NANOMETER'
        #     else:
        #         module_logger.warning(
        #             'Could not find microscope type to calculate theoretical resolution. Falling back into Wide-Field')
        #         theoretical_res = calculate_theoretcal_resolution('wf', na, refractive_index, emission_wave,
        #                                                           excitation_wave=excitation_wave)
        #
        #     return theoretical_res

        #
        # if refractive_index is None:
        #     module_logger.warning('Refractive index is being guessed. Nyquist criteria will not be correct.')
        #     if na > .8:
        #         refractive_index = 1.5
        #     else:
        #         refractive_index = 1.0
        # alpha = asin(na / refractive_index)
        # # Theoretical resolutions for confocal microscope are only attained with very closed pinhole
        # # We add a tolerance factor to render theoretical resolution more realistic
        # tolerance = 1.6
        # nyquist_delta = {}
        # if microscope_type is None:
        #     module_logger.warning(
        #         'Microscope type undefined to calculate Nyquist criterion. Falling back into Wide-Field')
        #     nyquist_delta = calculate_nyquist('wf', na, refractive_index, emission_wave, excitation_wave)
        # elif microscope_type.lower() in ['wf', 'wide-field', 'widefield']:
        #     nyquist_delta['lateral'] = emission_wave / (4 * refractive_index * sin(alpha))
        #     nyquist_delta['axial'] = emission_wave / (2 * refractive_index * (1 - cos(alpha)))
        #     nyquist_delta['units'] = 'NANOMETER'
        # elif microscope_type.lower() in ['confocal']:
        #     nyquist_delta['lateral'] = tolerance * excitation_wave / (8 * refractive_index * sin(alpha))
        #     nyquist_delta['axial'] = tolerance * excitation_wave / (4 * refractive_index * (1 - cos(alpha)))
        #     nyquist_delta['units'] = 'NANOMETER'
        # else:
        #     module_logger.warning(
        #         'Could not find microscope type to calculate Nyquist criterion. Falling back into Wide-Field')
        #     nyquist_delta = calculate_nyquist('wf', na, refractive_index, emission_wave, excitation_wave)
        #
        # return nyquist_delta



from datetime import datetime

# Import sample superclass
from microscopemetrics.samples.samples import Analysis, Configurator

# Creating logging services
import logging

module_logger = logging.getLogger("metrics.samples.dataset")


class DatasetConfigurator(Configurator):
    """This class handles the configuration properties of the excitation_power sample
    - Defines configuration properties
    - Helps in the generation of analysis_config files"""

    CONFIG_SECTION = "DATASET"
    ANALYSES = ["laser_power_measurement"]

    def __init__(self, config):
        super().__init__(config)


@DatasetConfigurator.register_sample_analysis
class DatasetAnalysis(Analysis):
    """This class handles the Excitation_power sample:
    - Defines the logic of the associated analyses
    - Defines the creation of reports"""

    def __init__(self, config=None):
        dataset_analysis_to_func = {
            "laser_power_measurement": self.analyze_laser_power_measurement,
        }
        self.configurator = DatasetConfigurator(config)
        super().__init__(
            config=config, dataset_analysis_to_func=dataset_analysis_to_func
        )

    @staticmethod
    def analyze_laser_power_measurement(dataset, config):
        """Opens a Map annotation where the user can enter the power measurements of the Excitation lasers. Those shoudl
        be blocked upon validation

        :param dataset: dataset instance
        :param config: MetricsConfig instance defining analysis configuration.
                       Must contain the analysis parameters defined by the configurator

        :returns a list of images
                 a list of rois
                 a list of tags
                 a list of dicts
                 a dict containing table_names and tables
        """
        module_logger.info(f"Generating light sources annotation...")

        wavelengths = config.getlistint("laser_power_measurement_wavelengths", None)
        power_units = config.get("laser_power_measurement_units")
        measurement_method = config.get("laser_power_measurement_method")

        if wavelengths is None:
            module_logger.error(
                "Config defined to do excitation power measurements but No wavelengths were provided"
            )
            wavelengths = list()

        key_values = {"analysis_date": datetime.now().strftime("%Y-%m-%d")}
        for wave in wavelengths:
            key_values.update({str(wave): ""})
        key_values["power_units"] = power_units
        key_values["measurement_method"] = measurement_method

        out_images = []
        out_rois = []
        out_tags = []
        out_dicts = [key_values]
        out_editables = [True]
        out_tables = []

        return out_images, out_rois, out_tags, out_dicts, out_editables, out_tables



# This is a place to hold mere utilities for metrics

from configparser import ConfigParser
import json
from scipy import special
import numpy as np


## Some useful functions
def convert_SI(val, unit_in, unit_out):
    si = {
        "nanometer": 0.000000001,
        "micrometer": 0.000001,
        "millimeter": 0.001,
        "meter": 1.0,
    }
    return val * si[unit_in.lower()] / si[unit_out.lower()]


# def airy_fun(x, centre, a, exp):  # , amp, bg):
#     if (x - centre) == 0:
#         return a * .5 ** exp
#     else:
#         return a * (special.j1(x - centre) / (x - centre)) ** exp
#
#
# def multi_airy_fun(x, *params):
#     y = np.zeros_like(x)
#     for i in range(0, len(params), 3):
#         y = y + airy_fun(x, params[i], params[i+1], params[i+2])
#     return y
def airy_fun(x, centre, amp):  # , exp):  # , amp, bg):
    with np.errstate(divide="ignore", invalid="ignore"):
        return np.where(
            (x - centre) == 0,
            amp * 0.5**2,
            amp * (special.j1(x - centre) / (x - centre)) ** 2,
        )


def gaussian_fun(x, background, amplitude, center, sd):
    gauss = np.exp(-np.power(x - center, 2.0) / (2 * np.power(sd, 2.0)))
    return background + (amplitude - background) * gauss


def multi_airy_fun(x, *params):
    y = np.zeros_like(x)
    for i in range(0, len(params), 2):
        y = y + airy_fun(x, params[i], params[i + 1])
    return y


def wavelength_to_rgb(wavelength, gamma=0.8):

    """
    Copied from https://www.noah.org/wiki/Wavelength_to_RGB_in_Python
    This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html
    """

    wavelength = float(wavelength)
    if 380 < wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        r = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        g = 0.0
        b = (1.0 * attenuation) ** gamma
    elif 440 < wavelength <= 490:
        r = 0.0
        g = ((wavelength - 440) / (490 - 440)) ** gamma
        b = 1.0
    elif 490 < wavelength <= 510:
        r = 0.0
        g = 1.0
        b = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif 510 < wavelength <= 580:
        r = ((wavelength - 510) / (580 - 510)) ** gamma
        g = 1.0
        b = 0.0
    elif 580 < wavelength <= 645:
        r = 1.0
        g = (-(wavelength - 645) / (645 - 580)) ** gamma
        b = 0.0
    elif 645 < wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        r = (1.0 * attenuation) ** gamma
        g = 0.0
        b = 0.0
    else:
        r = 0.0
        g = 0.0
        b = 0.0
    r *= 255
    g *= 255
    b *= 255
    return int(r), int(g), int(b)


class MetricsConfig(ConfigParser):
    def getjson(self, section, option, **kwargs):
        value = self.get(section, option, **kwargs)
        try:
            output = json.loads(value)
        except json.JSONDecodeError as e:
            raise e
        return output

    def getlist(self, section, option, **kwargs):
        output = self.getjson(section, option, **kwargs)

        if type(output) is list:
            return output
        else:
            raise Exception(
                f'The config option "{option}" in section "{section}" is not formatted as a list'
            )

    def getlistint(self, section, option, **kwargs):
        try:
            output = [int(x) for x in self.getlist(section, option, **kwargs)]
            return output
        except Exception as e:
            print(
                f'Some element in config option "{option}" in section "{section}" cannot be coerced into a integer'
            )
            raise e

    def getlistfloat(self, section, option, **kwargs):
        try:
            output = [float(x) for x in self.getlist(section, option, **kwargs)]
            return output
        except Exception as e:
            print(
                f'Some element in config option "{option}" in section "{section}" cannot be coerced into a float'
            )
            raise e



# Import sample infrastructure
from itertools import product

from microscopemetrics.samples import *

from typing import Union, Tuple, List

# Import analysis tools
import numpy as np
from pandas import DataFrame
from skimage.transform import hough_line  # hough_line_peaks, probabilistic_hough_line
from scipy.signal import find_peaks
from scipy.optimize import curve_fit
from scipy.interpolate import griddata
from microscopemetrics.analysis.tools import (
    segment_image,
    compute_distances_matrix,
    compute_spots_properties,
)
from ..utilities.utilities import multi_airy_fun, airy_fun


@register_image_analysis
class ArgolightBAnalysis(Analysis):
    """This class handles the analysis of the Argolight sample pattern B"""

    def __init__(self):
        super().__init__(
            output_description="Analysis output of the 'SPOTS' matrix (pattern B) from the argolight sample. "
            "It contains chromatic shifts and homogeneity."
        )
        self.add_data_requirement(
            name="argolight_b",
            description="Input image in the form of a numpy array",
            data_type=np.ndarray,
        )
        self.add_metadata_requirement(
            name="spots_distance",
            description="Distance between argolight spots",
            data_type=float,
            units="MICRON",
            optional=False,
        )
        self.add_metadata_requirement(
            name="pixel_size",
            description="Physical size of the voxel in z, y and x",
            data_type=Tuple[float, float, float],
            units="MICRON",
            optional=False,
        )
        self.add_metadata_requirement(
            name="sigma",
            description="Smoothing factor for objects detection",
            data_type=Tuple[float, float, float],
            optional=True,
            default=(1, 3, 3),
        )
        self.add_metadata_requirement(
            name="lower_threshold_correction_factors",
            description="Correction factor for the lower thresholds. Must be a tuple with len = nr "
            "of channels or a float if all equal",
            data_type=Union[List[float], Tuple[float], float],
            optional=True,
            default=None,
        )
        self.add_metadata_requirement(
            name="upper_threshold_correction_factors",
            description="Correction factor for the upper thresholds. Must be a tuple with len = nr "
            "of channels or a float if all equal",
            data_type=Union[List[float], Tuple[float], float],
            optional=True,
            default=None,
        )
        self.add_metadata_requirement(
            name="remove_center_cross",
            description="Remove the center cross found in some Argolight patterns",
            data_type=bool,
            optional=True,
            default=False,
        )

    def run(self):
        logger.info("Validating requirements...")
        if not self.validate_requirements():
            logger.error("Metadata requirements are not valid")
            return False

        logger.info("Analyzing spots image...")

        # Calculating the distance between spots in pixels with a security margin
        min_distance = round(
            (self.get_metadata_values("spots_distance") * 0.3)
            / max(self.get_metadata_values("pixel_size")[-2:])
        )

        # Calculating the maximum tolerated distance in microns for the same spot in a different channels
        max_distance = self.get_metadata_values("spots_distance") * 0.4

        labels = segment_image(
            image=self.get_data_values("argolight_b"),
            min_distance=min_distance,
            sigma=self.get_metadata_values("sigma"),
            method="local_max",
            low_corr_factors=self.get_metadata_values(
                "lower_threshold_correction_factors"
            ),
            high_corr_factors=self.get_metadata_values(
                "upper_threshold_correction_factors"
            ),
        )

        self.output.append(
            model.Image(
                name=list(self.input.data.keys())[0],
                description="Labels image with detected spots. "
                "Image intensities correspond to roi labels.",
                data=labels,
            )
        )

        spots_properties, spots_positions = compute_spots_properties(
            image=self.get_data_values("argolight_b"),
            labels=labels,
            remove_center_cross=self.get_metadata_values("remove_center_cross"),
        )

        distances_df = compute_distances_matrix(
            positions=spots_positions,
            max_distance=max_distance,
            pixel_size=self.get_metadata_values("pixel_size"),
        )

        properties_kv = {}
        properties_df = DataFrame()

        for ch, ch_spot_props in enumerate(spots_properties):
            ch_df = DataFrame()
            ch_df["channel"] = [ch for _ in ch_spot_props]
            ch_df["mask_labels"] = [p["label"] for p in ch_spot_props]
            ch_df["volume"] = [p["area"] for p in ch_spot_props]
            ch_df["roi_volume_units"] = "VOXEL"
            ch_df["max_intensity"] = [p["max_intensity"] for p in ch_spot_props]
            ch_df["min_intensity"] = [p["min_intensity"] for p in ch_spot_props]
            ch_df["mean_intensity"] = [p["mean_intensity"] for p in ch_spot_props]
            ch_df["integrated_intensity"] = [
                p["integrated_intensity"] for p in ch_spot_props
            ]
            ch_df["z_weighted_centroid"] = [
                p["weighted_centroid"][0] for p in ch_spot_props
            ]
            ch_df["y_weighted_centroid"] = [
                p["weighted_centroid"][1] for p in ch_spot_props
            ]
            ch_df["x_weighted_centroid"] = [
                p["weighted_centroid"][2] for p in ch_spot_props
            ]
            ch_df["roi_weighted_centroid_units"] = "PIXEL"

            # Key metrics for spots intensities
            properties_kv[f"nr_of_spots_ch{ch:02d}"] = len(ch_df)
            properties_kv[f"max_intensity_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].max().item()
            )
            properties_kv[f"max_intensity_roi_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].argmax().item()
            )
            properties_kv[f"min_intensity_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].min().item()
            )
            properties_kv[f"min_intensity_roi_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].argmin().item()
            )
            properties_kv[f"mean_intensity_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].mean().item()
            )
            properties_kv[f"median_intensity_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].median().item()
            )
            properties_kv[f"std_mean_intensity_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].std().item()
            )
            properties_kv[f"mad_mean_intensity_ch{ch:02d}"] = (
                ch_df["integrated_intensity"].mad().item()
            )
            properties_kv[f"min-max_intensity_ratio_ch{ch:02d}"] = (
                properties_kv[f"min_intensity_ch{ch:02d}"]
                / properties_kv[f"max_intensity_ch{ch:02d}"]
            )

            properties_df = properties_df.append(ch_df)

            channel_shapes = [
                model.Point(
                    x=p["weighted_centroid"][2].item(),
                    y=p["weighted_centroid"][1].item(),
                    z=p["weighted_centroid"][0].item(),
                    c=ch,
                    label=f'{p["label"]}',
                )
                for p in ch_spot_props
            ]
            self.output.append(
                model.Roi(
                    name=f"Centroids_ch{ch:03d}",
                    description=f"weighted centroids channel {ch}",
                    shapes=channel_shapes,
                )
            )

        distances_kv = {"distance_units": self.get_metadata_units("pixel_size")}

        for a, b in product(
            distances_df.channel_a.unique(), distances_df.channel_b.unique()
        ):
            temp_df = distances_df[
                (distances_df.channel_a == a) & (distances_df.channel_b == b)
            ]
            a = int(a)
            b = int(b)

            distances_kv[
                f"mean_3d_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.dist_3d.mean().item()
            distances_kv[
                f"median_3d_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.dist_3d.median().item()
            distances_kv[
                f"std_3d_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.dist_3d.std().item()
            distances_kv[
                f"mad_3d_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.dist_3d.mad().item()
            distances_kv[
                f"mean_z_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.z_dist.mean().item()
            distances_kv[
                f"median_z_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.z_dist.median().item()
            distances_kv[
                f"std_z_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.z_dist.std().item()
            distances_kv[
                f"mad_z_dist_ch{a:02d}_ch{b:02d}"
            ] = temp_df.z_dist.mad().item()

        self.output.append(
            model.KeyValues(
                name="Intensity Key Annotations",
                description="Key Intensity Measurements on Argolight D spots",
                key_values=properties_kv,
            )
        )

        self.output.append(
            model.KeyValues(
                name="Distances Key Annotations",
                description="Key Distance Measurements on Argolight D spots",
                key_values=distances_kv,
            )
        )

        self.output.append(
            model.Table(
                name="Properties",
                description="Analysis_argolight_D_properties",
                table=properties_df,
            )
        )

        self.output.append(
            model.Table(
                name="Distances",
                description="Analysis_argolight_D_distances",
                table=distances_df,
            )
        )

        return True


@register_image_analysis
class ArgolightEAnalysis(Analysis):
    """This class handles the analysis of the Argolight sample pattern E with lines along the X or Y axis"""

    def __init__(self):
        super().__init__(
            output_description="Analysis output of the lines (pattern E) from the argolight sample. "
            "It contains resolution data on the axis indicated:"
            "- axis 1 = Y resolution = lines along X axis"
            "- axis 2 = X resolution = lines along Y axis"
        )
        self.add_data_requirement(
            name="argolight_e",
            description="Input image in the form of a numpy array",
            data_type=np.ndarray,
        )
        self.add_metadata_requirement(
            name="pixel_size",
            description="Physical size of the voxel in z, y and x",
            data_type=Tuple[float, float, float],
            units="MICRON",
            optional=False,
        )
        self.add_metadata_requirement(
            name="axis",
            description="axis along which resolution is being measured. 1=Y, 2=X",
            data_type=int,
            optional=False,
        )
        self.add_metadata_requirement(
            name="measured_band",
            description="Fraction of the image across which intensity profiles are measured",
            data_type=float,
            optional=True,
            default=0.4,
        )

    def run(self):
        """A intermediate function to specify the axis to be analyzed"""

        logger.info("Validating requirements...")
        if not self.validate_requirements():
            logger.error("Metadata requirements are not valid")
            return False

        logger.info("Analyzing resolution...")

        return self._analyze_resolution(
            image=self.get_data_values("argolight_e"),
            axis=self.get_metadata_values("axis"),
            measured_band=self.get_metadata_values("measured_band"),
            pixel_size=self.get_metadata_values("pixel_size"),
            pixel_size_units=self.get_metadata_units("pixel_size"),
        )

    def _analyze_resolution(
        self, image, axis, measured_band, pixel_size, pixel_size_units
    ):
        (
            profiles,
            z_planes,
            peak_positions,
            peak_heights,
            resolution_values,
            resolution_indexes,
            resolution_method,
        ) = _compute_resolution(
            image=image,
            axis=axis,
            measured_band=measured_band,
            prominence=0.264,
            do_angle_refinement=False,
        )
        # resolution in native units
        resolution_values = [x * pixel_size[axis] for x in resolution_values]

        key_values = {
            f"ch{ch:02d}_{resolution_method}_resolution": res.item()
            for ch, res in enumerate(resolution_values)
        }

        key_values["resolution_units"] = pixel_size_units
        key_values["resolution_axis"] = axis
        key_values["measured_band"] = measured_band

        for ch, indexes in enumerate(resolution_indexes):
            key_values[f"peak_positions_ch{ch:02d}"] = [
                (peak_positions[ch][ind].item(), peak_positions[ch][ind + 1].item())
                for ind in indexes
            ]
            key_values[f"peak_heights_ch{ch:02d}"] = [
                (peak_heights[ch][ind].item(), peak_heights[ch][ind + 1].item())
                for ind in indexes
            ]
            key_values[f"focus_ch{ch:02d}"] = z_planes[ch].item()

        out_tables = {}

        # Populate tables and rois
        for ch, profile in enumerate(profiles):
            out_tables.update(_profile_to_table(profile, ch))
            shapes = []
            for pos in key_values[f"peak_positions_ch{ch:02d}"]:
                for peak in pos:
                    # Measurements are taken at center of pixel so we add .5 pixel to peak positions
                    if axis == 1:  # Y resolution -> horizontal rois
                        axis_len = image.shape[-2]
                        x1_pos = int((axis_len / 2) - (axis_len * measured_band / 2))
                        y1_pos = peak + 0.5
                        x2_pos = int((axis_len / 2) + (axis_len * measured_band / 2))
                        y2_pos = peak + 0.5
                    elif axis == 2:  # X resolution -> vertical rois
                        axis_len = image.shape[-1]
                        y1_pos = int((axis_len / 2) - (axis_len * measured_band / 2))
                        x1_pos = peak + 0.5
                        y2_pos = int((axis_len / 2) + (axis_len * measured_band / 2))
                        x2_pos = peak + 0.5

                    shapes.append(
                        model.Line(
                            x1=x1_pos,
                            y1=y1_pos,
                            x2=x2_pos,
                            y2=y2_pos,
                            z=z_planes[ch],
                            c=ch,
                        )
                    )

            self.output.append(
                model.Roi(
                    name=f"Peaks_ch{ch:03d}",
                    description=f"Lines where highest Rayleigh resolution was found in channel {ch}",
                    shapes=shapes,
                )
            )
        self.output.append(
            model.KeyValues(
                name="Key-Value Annotations",
                description=f"Measurements on Argolight E pattern along axis={axis}",
                key_values=key_values,
            )
        )
        self.output.append(
            model.Table(
                name="Profiles",
                description="Raw and fitted profiles across the center of the image along the "
                "defined axis",
                table=DataFrame.from_dict(out_tables),
            )
        )

        return True


class ArgolightReporter(Reporter):
    """Reporter subclass to produce Argolight sample figures"""

    def __init__(self):
        image_report_to_func = {
            "spots": self.full_report_spots,
            "vertical_resolution": self.full_report_vertical_resolution,
            "horizontal_resolution": self.full_report_horizontal_resolution,
        }

        super().__init__(image_report_to_func=image_report_to_func)

    def produce_image_report(self, image):
        pass

    def full_report_spots(self, image):
        pass

    def full_report_vertical_resolution(self, image):
        pass

    def full_report_horizontal_resolution(self, image):
        pass

    def plot_homogeneity_map(self, image):
        nr_channels = image.getSizeC()
        x_dim = image.getSizeX()
        y_dim = image.getSizeY()

        tables = self.get_tables(
            image, namespace_start="metrics", name_filter="properties"
        )
        if len(tables) != 1:
            raise Exception(
                "There are none or more than one properties tables. Verify data integrity."
            )
        table = tables[0]

        row_count = table.getNumberOfRows()
        col_names = [c.name for c in table.getHeaders()]
        wanted_columns = [
            "channel",
            "max_intensity",
            "mean_intensity",
            "integrated_intensity",
            "x_weighted_centroid",
            "y_weighted_centroid",
        ]

        fig, axes = plt.subplots(
            ncols=nr_channels, nrows=3, squeeze=False, figsize=(3 * nr_channels, 9)
        )

        for ch in range(nr_channels):
            data = table.slice(
                [col_names.index(w_col) for w_col in wanted_columns],
                table.getWhereList(
                    condition=f"channel=={ch}",
                    variables={},
                    start=0,
                    stop=row_count,
                    step=0,
                ),
            )
            max_intensity = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "max_intensity"
                ]
            )
            integrated_intensity = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "integrated_intensity"
                ]
            )
            x_positions = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "x_weighted_centroid"
                ]
            )
            y_positions = np.array(
                [
                    val
                    for col in data.columns
                    for val in col.values
                    if col.name == "y_weighted_centroid"
                ]
            )
            grid_x, grid_y = np.mgrid[0:x_dim, 0:y_dim]
            image_intensities = get_intensities(image, c_range=ch, t_range=0).max(0)

            try:
                interpolated_max_int = griddata(
                    np.stack((x_positions, y_positions), axis=1),
                    max_intensity,
                    (grid_x, grid_y),
                    method="linear",
                )
                interpolated_intgr_int = griddata(
                    np.stack((x_positions, y_positions), axis=1),
                    integrated_intensity,
                    (grid_x, grid_y),
                    method="linear",
                )
            except Exception as e:
                # TODO: Log a warning
                interpolated_max_int = np.zeros((256, 256))

            ax = axes.ravel()
            ax[ch] = plt.subplot(3, 4, ch + 1)

            ax[ch].imshow(np.squeeze(image_intensities), cmap="gray")
            ax[ch].set_title("MIP_" + str(ch))

            ax[ch + nr_channels].imshow(
                np.flipud(interpolated_intgr_int),
                extent=(0, x_dim, y_dim, 0),
                origin="lower",
                cmap=cm.hot,
                vmin=np.amin(integrated_intensity),
                vmax=np.amax(integrated_intensity),
            )
            ax[ch + nr_channels].plot(x_positions, y_positions, "k.", ms=2)
            ax[ch + nr_channels].set_title("Integrated_int_" + str(ch))

            ax[ch + 2 * nr_channels].imshow(
                np.flipud(interpolated_max_int),
                extent=(0, x_dim, y_dim, 0),
                origin="lower",
                cmap=cm.hot,
                vmin=np.amin(image_intensities),
                vmax=np.amax(image_intensities),
            )
            ax[ch + 2 * nr_channels].plot(x_positions, y_positions, "k.", ms=2)
            ax[ch + 2 * nr_channels].set_title("Max_int_" + str(ch))

        plt.show()

    def plot_distances_map(self, image):
        nr_channels = image.getSizeC()
        x_dim = image.getSizeX()
        y_dim = image.getSizeY()

        tables = get_tables(image, namespace_start="metrics", name_filter="distances")
        if len(tables) != 1:
            raise Exception(
                "There are none or more than one distances tables. Verify data integrity."
            )
        table = tables[0]
        row_count = table.getNumberOfRows()
        col_names = [c.name for c in table.getHeaders()]

        # We need the positions too
        pos_tables = get_tables(
            image, namespace_start="metrics", name_filter="properties"
        )
        if len(tables) != 1:
            raise Exception(
                "There are none or more than one positions tables. Verify data integrity."
            )
        pos_table = pos_tables[0]
        pos_row_count = pos_table.getNumberOfRows()
        pos_col_names = [c.name for c in pos_table.getHeaders()]

        fig, axes = plt.subplots(
            ncols=nr_channels - 1,
            nrows=nr_channels,
            squeeze=False,
            figsize=((nr_channels - 1) * 3, nr_channels * 3),
        )

        ax_index = 0
        for ch_A in range(nr_channels):
            pos_data = pos_table.slice(
                [
                    pos_col_names.index(w_col)
                    for w_col in [
                        "channel",
                        "mask_labels",
                        "x_weighted_centroid",
                        "y_weighted_centroid",
                    ]
                ],
                pos_table.getWhereList(
                    condition=f"channel=={ch_A}",
                    variables={},
                    start=0,
                    stop=pos_row_count,
                    step=0,
                ),
            )

            mask_labels = np.array(
                [
                    val
                    for col in pos_data.columns
                    for val in col.values
                    if col.name == "mask_labels"
                ]
            )
            x_positions = np.array(
                [
                    val
                    for col in pos_data.columns
                    for val in col.values
                    if col.name == "x_weighted_centroid"
                ]
            )
            y_positions = np.array(
                [
                    val
                    for col in pos_data.columns
                    for val in col.values
                    if col.name == "y_weighted_centroid"
                ]
            )
            positions_map = np.stack((x_positions, y_positions), axis=1)

            for ch_B in [i for i in range(nr_channels) if i != ch_A]:
                data = table.slice(
                    list(range(len(col_names))),
                    table.getWhereList(
                        condition=f"(channel_A=={ch_A})&(channel_B=={ch_B})",
                        variables={},
                        start=0,
                        stop=row_count,
                        step=0,
                    ),
                )
                labels_map = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "ch_A_roi_labels"
                    ]
                )
                labels_map += 1  # Mask labels are augmented by one as 0 is background
                distances_map_3d = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_3d"
                    ]
                )
                distances_map_x = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_x"
                    ]
                )
                distances_map_y = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_y"
                    ]
                )
                distances_map_z = np.array(
                    [
                        val
                        for col in data.columns
                        for val in col.values
                        if col.name == "distance_z"
                    ]
                )

                filtered_positions = positions_map[
                    np.intersect1d(
                        mask_labels, labels_map, assume_unique=True, return_indices=True
                    )[1],
                    :,
                ]

                grid_x, grid_y = np.mgrid[0:x_dim:1, 0:y_dim:1]
                interpolated = griddata(
                    filtered_positions,
                    distances_map_3d,
                    (grid_x, grid_y),
                    method="cubic",
                )

                ax = axes.ravel()
                ax[ax_index].imshow(
                    np.flipud(interpolated),
                    extent=(0, x_dim, y_dim, 0),
                    origin="lower",
                    cmap=cm.hot,
                    vmin=np.amin(distances_map_3d),
                    vmax=np.amax(distances_map_3d),
                )
                ax[ax_index].set_title(f"Distance Ch{ch_A}-Ch{ch_B}")

                ax_index += 1

        plt.show()


def _profile_to_table(profile, channel):
    table = {f"raw_profile_ch{channel:02d}": [v.item() for v in profile[0, :]]}

    for p in range(1, profile.shape[0]):
        table.update(
            {
                f"fitted_profile_ch{channel:03d}_peak{p:03d}": [
                    v.item() for v in profile[p, :]
                ]
            }
        )

    return table


def _fit(
    profile, peaks_guess, amp=4, exp=2, lower_amp=3, upper_amp=5, center_tolerance=1
):
    guess = []
    lower_bounds = []
    upper_bounds = []
    for p in peaks_guess:
        guess.append(p)  # peak center
        guess.append(amp)  # peak amplitude
        lower_bounds.append(p - center_tolerance)
        lower_bounds.append(lower_amp)
        upper_bounds.append(p + center_tolerance)
        upper_bounds.append(upper_amp)

    x = np.linspace(0, profile.shape[0], profile.shape[0], endpoint=False)

    popt, pcov = curve_fit(
        multi_airy_fun, x, profile, p0=guess, bounds=(lower_bounds, upper_bounds)
    )

    opt_peaks = popt[::2]
    # opt_amps = [a / 4 for a in popt[1::2]]  # We normalize back the amplitudes to the unity
    opt_amps = popt[1::2]

    fitted_profiles = np.zeros((len(peaks_guess), profile.shape[0]))
    for i, (c, a) in enumerate(zip(opt_peaks, opt_amps)):
        fitted_profiles[i, :] = airy_fun(x, c, a)

    return opt_peaks, opt_amps, fitted_profiles


def _compute_channel_resolution(
    channel, axis, prominence, measured_band, do_fitting=True, do_angle_refinement=False
):
    """Computes the resolution on a pattern of lines with increasing separation"""
    # find the most contrasted z-slice
    z_stdev = np.std(channel, axis=(1, 2))
    z_focus = np.argmax(z_stdev)
    focus_slice = channel[z_focus]

    # TODO: verify angle and correct
    if do_angle_refinement:
        # Set a precision of 0.1 degree.
        tested_angles = np.linspace(-np.pi / 2, np.pi / 2, 1800)
        h, theta, d = hough_line(focus_slice, theta=tested_angles)

    # Cut a band of that found peak
    # Best we can do now is just to cut a band in the center
    # We create a profiles along which we average signal
    axis_len = focus_slice.shape[-axis]
    weight_profile = np.zeros(axis_len)
    # Calculates a band of relative width 'image_fraction' to integrate the profile
    weight_profile[
        int((axis_len / 2) - (axis_len * measured_band / 2)) : int(
            (axis_len / 2) + (axis_len * measured_band / 2)
        )
    ] = 1
    profile = np.average(focus_slice, axis=-axis, weights=weight_profile)

    normalized_profile = (profile - np.min(profile)) / np.ptp(profile)

    # Find peaks: We implement Rayleigh limits that will be refined downstream
    peak_positions, properties = find_peaks(
        normalized_profile,
        height=0.3,
        distance=2,
        prominence=prominence / 4,
    )

    # From the properties we are interested in the amplitude
    # peak_heights = [h for h in properties['peak_heights']]
    ray_filtered_peak_pos = []
    ray_filtered_peak_heights = []

    for peak, height, prom in zip(
        peak_positions, properties["peak_heights"], properties["prominences"]
    ):
        if (
            prom / height
        ) > prominence:  # This is calculating the prominence in relation to the local intensity
            ray_filtered_peak_pos.append(peak)
            ray_filtered_peak_heights.append(height)

    peak_positions = ray_filtered_peak_pos
    peak_heights = ray_filtered_peak_heights

    if do_fitting:
        peak_positions, peak_heights, fitted_profiles = _fit(
            normalized_profile, peak_positions
        )
        normalized_profile = np.append(
            np.expand_dims(normalized_profile, 0), fitted_profiles, axis=0
        )

    # Find the closest peaks to return it as a measure of resolution
    peaks_distances = [
        abs(a - b) for a, b in zip(peak_positions[0:-2], peak_positions[1:-1])
    ]
    res = min(peaks_distances)  # TODO: capture here the case where there are no peaks!
    res_indices = [i for i, x in enumerate(peaks_distances) if x == res]

    return normalized_profile, z_focus, peak_positions, peak_heights, res, res_indices


def _compute_resolution(
    image, axis, measured_band, prominence, do_angle_refinement=False
):
    profiles = list()
    z_planes = []
    peaks_positions = list()
    peaks_heights = []
    resolution_values = []
    resolution_indexes = []
    resolution_method = "rayleigh"

    for c in range(image.shape[1]):  # TODO: Deal with Time here
        prof, zp, pk_pos, pk_heights, res, res_ind = _compute_channel_resolution(
            channel=np.squeeze(image[:, c, ...]),
            axis=axis,
            prominence=prominence,
            measured_band=measured_band,
            do_angle_refinement=do_angle_refinement,
        )
        profiles.append(prof)
        z_planes.append(zp)
        peaks_positions.append(pk_pos)
        peaks_heights.append(pk_heights)
        resolution_values.append(res)
        resolution_indexes.append(res_ind)

    return (
        profiles,
        z_planes,
        peaks_positions,
        peaks_heights,
        resolution_values,
        resolution_indexes,
        resolution_method,
    )


# Calculate 2D FFT
# slice_2d = raw_img[17, ...].reshape([1, n_channels, x_size, y_size])
# fft_2D = fft_2d(slice_2d)

# Calculate 3D FFT
# fft_3D = fft_3d(spots_image)
#
# plt.imshow(np.log(fft_3D[2, :, :, 1]))  # , cmap='hot')
# # plt.imshow(np.log(fft_3D[2, 23, :, :]))  # , cmap='hot')
# plt.show()
#



from typing import Tuple

import numpy as np
from pandas import DataFrame
from skimage.filters import gaussian
from skimage.feature import peak_local_max
from scipy.optimize import curve_fit, fsolve
from pydantic.color import Color

from ..utilities.utilities import airy_fun, gaussian_fun

# Import sample superclass
from microscopemetrics.samples import *

# Creating logging services
import logging

module_logger = logging.getLogger("metrics.samples.psf_beads")


def _fit_gaussian(profile, guess=None):
    if guess is None:
        guess = [profile.min(), profile.max(), profile.argmax(), 0.8]
    x = np.linspace(0, profile.shape[0], profile.shape[0], endpoint=False)
    popt, pcov = curve_fit(gaussian_fun, x, profile, guess)

    fitted_profile = gaussian_fun(x, popt[0], popt[1], popt[2], popt[3])
    fwhm = popt[3] * 2.35482

    return fitted_profile, fwhm


def _fit_airy(profile, guess=None):
    if guess is None:
        guess = [profile.argmax(), 4 * profile.max()]
    x = np.linspace(0, profile.shape[0], profile.shape[0], endpoint=False)
    popt, pcov = curve_fit(airy_fun, x, profile, guess)

    fitted_profile = airy_fun(x, popt[0], popt[1])

    def _f(d):
        return (
            airy_fun(d, popt[0], popt[1])
            - (fitted_profile.max() - fitted_profile.min()) / 2
        )

    guess = np.array([fitted_profile.argmax() - 1, fitted_profile.argmax() + 1])
    v = fsolve(_f, guess)
    fwhm = abs(v[1] - v[0])

    return fitted_profile, fwhm


@register_image_analysis
class PSFBeadsAnalysis(Analysis):
    """This class handles a PSF beads sample"""

    def __init__(self, config=None):
        super().__init__(
            output_description="Analysis output of samples containing PSF grade fluorescent beads. "
            "It contains information about resolution."
        )
        self.add_data_requirement(
            name="beads_image",
            description="The image containing the beads in teh form of a numpy array",
            data_type=np.ndarray,
        )
        self.add_metadata_requirement(
            name="pixel_size",
            description="Physical size of the voxel in z, y and x",
            data_type=Tuple[float, float, float],
            units="MICRON",
            optional=False,
        )
        self.add_metadata_requirement(
            name="min_lateral_distance_factor",
            description="Minimal distance that has to separate laterally the beads represented as the "
            "number of times the theoretical resolution.",
            data_type=int,
            optional=True,
            default=20,
        )
        self.add_metadata_requirement(
            name="theoretical_fwhm_lateral_res",
            description="Theoretical FWHM lateral resolution of the sample.",
            data_type=float,
            units="MICRON",
            optional=False,
        )
        self.add_metadata_requirement(
            name="theoretical_fwhm_axial_res",
            description="Theoretical FWHM axial resolution of the sample.",
            data_type=float,
            units="MICRON",
            optional=False,
        )
        self.add_metadata_requirement(
            name="sigma",
            description="When provided, smoothing sigma to be applied to image prior to bead detection."
            "Does not apply to resolution measurements",
            data_type=float,
            optional=True,
            default=None,
        )

    @staticmethod
    def _analyze_bead(image):
        # Find the strongest sections to generate profiles
        x_max = np.max(image, axis=(0, 1))
        x_focus = np.argmax(x_max)
        y_max = np.max(image, axis=(0, 2))
        y_focus = np.argmax(y_max)
        z_max = np.max(image, axis=(1, 2))
        z_focus = np.argmax(z_max)

        # Generate profiles
        x_profile = np.squeeze(image[z_focus, y_focus, :])
        y_profile = np.squeeze(image[z_focus, :, x_focus])
        z_profile = np.squeeze(image[:, y_focus, x_focus])

        # Fitting the profiles
        x_fitted_profile, x_fwhm = _fit_airy(x_profile)
        y_fitted_profile, y_fwhm = _fit_airy(y_profile)
        z_fitted_profile, z_fwhm = _fit_airy(z_profile)

        return (
            (z_profile, y_profile, x_profile),
            (z_fitted_profile, y_fitted_profile, x_fitted_profile),
            (z_fwhm, y_fwhm, x_fwhm),
        )

    @staticmethod
    def _find_beads(image, min_distance, sigma=None):

        image = np.squeeze(image)
        image_mip = np.max(image, axis=0)

        if sigma is not None:
            image_mip = gaussian(
                image=image_mip, multichannel=False, sigma=sigma, preserve_range=True
            )

        # Find bead centers
        positions_2d = peak_local_max(
            image=image_mip, threshold_rel=0.2, min_distance=5, indices=True
        )
        # Add the mas intensity value in z
        positions_3d = np.insert(
            positions_2d[:],
            0,
            np.argmax(image[:, positions_2d[:, 0], positions_2d[:, 1]], axis=0),
            axis=1,
        )

        nr_beads = positions_2d.shape[0]
        module_logger.info(f"Beads found: {nr_beads}")

        # Exclude beads too close to the edge
        edge_keep_mask = (
            (positions_2d[:, 0] > min_distance)
            & (positions_2d[:, 0] < image_mip.shape[0] - min_distance)
            & (positions_2d[:, 1] > min_distance)
            & (positions_2d[:, 1] < image_mip.shape[1] - min_distance)
        )
        module_logger.info(
            f"Beads too close to the edge: {nr_beads - np.sum(edge_keep_mask)}"
        )

        # Exclude beads too close to each other
        proximity_keep_mask = np.ones((nr_beads, nr_beads), dtype=bool)
        for i, pos in enumerate(positions_2d):
            proximity_keep_mask[i] = (
                abs(positions_2d[:, 0] - pos[0]) > min_distance
            ) | (abs(positions_2d[:, 1] - pos[1]) > min_distance)
            proximity_keep_mask[i, i] = True  # Correcting the diagonal
        proximity_keep_mask = np.all(proximity_keep_mask, axis=0)
        module_logger.info(
            f"Beads too close to each other: {nr_beads - np.sum(proximity_keep_mask)}"
        )

        # Exclude beads too intense or too weak
        intensity_keep_mask = np.ones(nr_beads, dtype=bool)
        # TODO: Implement beads intensity filter
        module_logger.info(
            f"Beads too intense (probably more than one bead): {nr_beads - np.sum(intensity_keep_mask)}"
        )

        keep_mask = edge_keep_mask & proximity_keep_mask & intensity_keep_mask
        module_logger.info(f"Beads kept for analysis: {np.sum(keep_mask)}")

        positions = positions_3d[keep_mask, :]
        pos_edge_disc = positions_3d[np.logical_not(edge_keep_mask), :]
        pos_proximity_disc = positions_3d[np.logical_not(proximity_keep_mask), :]
        pos_intensity_disc = positions_3d[np.logical_not(intensity_keep_mask), :]

        bead_images = [
            image[
                :,
                (pos[1] - (min_distance // 2)) : (pos[1] + (min_distance // 2)),
                (pos[2] - (min_distance // 2)) : (pos[2] + (min_distance // 2)),
            ]
            for pos in positions
        ]
        return (
            bead_images,
            positions,
            pos_edge_disc,
            pos_proximity_disc,
            pos_intensity_disc,
        )

    def estimate_min_bead_distance(self):
        # TODO: get the resolution somewhere or pass it as a metadata
        res = 3  # theoretical resolution in pixels
        distance = self.get_metadata_values("min_lateral_distance_factor")
        return res * distance

    def run(self):
        """Analyzes images of sub-resolution beads in order to extract data on the optical
        performance of the microscope.
        """
        logger.info("Validating requirements...")
        if not self.validate_requirements():
            logger.error("Metadata requirements are not valid")
            return False

        logger.info("Analyzing spots image...")

        # Verify that image is not saturated
        if np.issubdtype(self.get_data_values("beads_image").dtype, np.integer):
            if (
                self.get_data_values("beads_image").max()
                == np.iinfo(self.get_data_values("beads_image").dtype).max
            ):
                logger.error(
                    "Image is saturated. No attempt to find beads will be done."
                )
                return False
        elif np.issubdtype(self.get_data_values("beads_image").dtype, np.float):
            if (
                self.get_data_values("beads_image").max()
                == np.finfo(self.get_data_values("beads_image").dtype).max
            ):
                logger.error(
                    "Image is saturated. No attempt to find beads will be done."
                )
                return False

        # Get some analysis_config parameters
        pixel_size_units = self.get_metadata_units("pixel_size")
        pixel_size = self.get_metadata_values("pixel_size")
        min_bead_distance = self.estimate_min_bead_distance()

        # Remove all negative intensities. eg. 3D-SIM images may contain negative values.
        image_data = np.clip(self.get_data_values("beads_image"), a_min=0, a_max=None)

        # Validating nyquist
        try:
            if pixel_size[1] > (
                2 * self.get_metadata_values("theoretical_fwhm_lateral_res")
            ):
                module_logger.warning(
                    "Nyquist criterion is not fulfilled in the lateral direction"
                )
            if pixel_size[0] > (
                2 * self.get_metadata_values("theoretical_fwhm_axial_res")
            ):
                module_logger.warning(
                    "Nyquist criterion is not fulfilled in the axial direction"
                )
        except (TypeError, IndexError) as e:
            module_logger.error("Could not validate Nyquist sampling criterion")

        (
            bead_images,
            positions,
            positions_edge_discarded,
            positions_proximity_discarded,
            positions_intensity_discarded,
        ) = self._find_beads(
            image=image_data,
            min_distance=min_bead_distance,
            sigma=self.get_metadata_values("sigma"),
        )

        for i, bead_image in enumerate(bead_images):
            self.output.append(
                model.Image(
                    name=f"bead_nr{i:02d}",
                    description=f"PSF bead crop for bead nr {i}",
                    data=np.expand_dims(bead_image, axis=(1, 2)),
                )
            )

        for i, position in enumerate(positions):
            self.output.append(
                model.Roi(
                    name=f"bead_nr{i:02d}_centroid",
                    description=f"Weighted centroid of bead nr {i}",
                    shapes=[
                        model.Point(
                            z=position[0].item(),
                            y=position[1].item(),
                            x=position[2].item(),
                            stroke_color=Color((0, 255, 0, 0.0)),
                            fill_color=Color((50, 255, 50, 0.1)),
                        )
                    ],
                )
            )

        edge_points = [
            model.Point(
                z=pos[0].item(),
                y=pos[1].item(),
                x=pos[2].item(),
                stroke_color=Color((255, 0, 0, 0.6)),
                fill_color=Color((255, 50, 50, 0.1)),
            )
            for pos in positions_edge_discarded
        ]
        self.output.append(
            model.Roi(
                name="Discarded_edge",
                description="Beads discarded for being to close to the edge of the image",
                shapes=edge_points,
            )
        )

        proximity_points = [
            model.Point(
                z=pos[0].item(),
                y=pos[1].item(),
                x=pos[2].item(),
                stroke_color=Color((255, 0, 0, 0.6)),
                fill_color=Color((255, 50, 50, 0.1)),
            )
            for pos in positions_proximity_discarded
        ]
        self.output.append(
            model.Roi(
                name="Discarded_proximity",
                description="Beads discarded for being to close to each other",
                shapes=proximity_points,
            )
        )

        intensity_points = [
            model.Point(
                z=pos[0].item(),
                y=pos[1].item(),
                x=pos[2].item(),
                stroke_color=Color((255, 0, 0, 0.6)),
                fill_color=Color((255, 50, 50, 0.1)),
            )
            for pos in positions_intensity_discarded
        ]
        self.output.append(
            model.Roi(
                name="Discarded_intensity",
                description="Beads discarded for being to intense or to weak. "
                "Suspected not being single beads",
                shapes=intensity_points,
            )
        )

        # Generate profiles and measure FWHM
        raw_profiles = []
        fitted_profiles = []
        fwhm_values = []
        for bead_image in bead_images:
            opr, fpr, fwhm = self._analyze_bead(bead_image)
            raw_profiles.append(opr)
            fitted_profiles.append(fpr)
            fwhm = tuple(f * ps for f, ps in zip(fwhm, pixel_size))
            fwhm_values.append(fwhm)

        properties_df = DataFrame()
        properties_df["bead_nr"] = range(len(bead_images))
        properties_df["max_intensity"] = [e.max() for e in bead_images]
        properties_df["min_intensity"] = [e.min() for e in bead_images]
        properties_df["z_centroid"] = [e[0] for e in positions]
        properties_df["y_centroid"] = [e[1] for e in positions]
        properties_df["x_centroid"] = [e[2] for e in positions]
        properties_df["centroid_units"] = "PIXEL"
        properties_df["z_fwhm"] = [e[0] for e in fwhm_values]
        properties_df["y_fwhm"] = [e[1] for e in fwhm_values]
        properties_df["x_fwhm"] = [e[2] for e in fwhm_values]
        properties_df["fwhm_units"] = pixel_size_units

        self.output.append(
            model.Table(
                name="Analysis_PSF_properties",
                description="Properties associated with the analysis",
                table=properties_df,
            )
        )

        profiles_z_df = DataFrame()
        profiles_y_df = DataFrame()
        profiles_x_df = DataFrame()

        for i, (raw_profile, fitted_profile) in enumerate(
            zip(raw_profiles, fitted_profiles)
        ):
            profiles_z_df[f"raw_z_profile_bead_{i:02d}"] = raw_profile[0]
            profiles_z_df[f"fitted_z_profile_bead_{i:02d}"] = fitted_profile[0]
            profiles_y_df[f"raw_y_profile_bead_{i:02d}"] = raw_profile[1]
            profiles_y_df[f"fitted_y_profile_bead_{i:02d}"] = fitted_profile[1]
            profiles_x_df[f"raw_x_profile_bead_{i:02d}"] = raw_profile[2]
            profiles_x_df[f"fitted_x_profile_bead_{i:02d}"] = fitted_profile[2]

        self.output.append(
            model.Table(
                name="Analysis_PSF_Z_profiles",
                description="Raw and fitted profiles along Z axis of beads",
                table=profiles_z_df,
            )
        )

        self.output.append(
            model.Table(
                name="Analysis_PSF_Y_profiles",
                description="Raw and fitted profiles along Y axis of beads",
                table=profiles_y_df,
            )
        )

        self.output.append(
            model.Table(
                name="Analysis_PSF_X_profiles",
                description="Raw and fitted profiles along X axis of beads",
                table=profiles_x_df,
            )
        )

        key_values = {"nr_of_beads_analyzed": positions.shape[0]}

        if key_values["nr_of_beads_analyzed"] == 0:
            key_values["resolution_mean_fwhm_z"] = "None"
            key_values["resolution_mean_fwhm_y"] = "None"
            key_values["resolution_mean_fwhm_x"] = "None"
            key_values["resolution_mean_fwhm_units"] = "None"
        else:
            key_values["resolution_mean_fwhm_z"] = properties_df["z_fwhm"].mean()
            key_values["resolution_median_fwhm_z"] = properties_df["z_fwhm"].median()
            key_values["resolution_stdev_fwhm_z"] = properties_df["z_fwhm"].std()
            key_values["resolution_mean_fwhm_y"] = properties_df["y_fwhm"].mean()
            key_values["resolution_median_fwhm_y"] = properties_df["y_fwhm"].median()
            key_values["resolution_stdev_fwhm_y"] = properties_df["y_fwhm"].std()
            key_values["resolution_mean_fwhm_x"] = properties_df["x_fwhm"].mean()
            key_values["resolution_median_fwhm_x"] = properties_df["x_fwhm"].median()
            key_values["resolution_stdev_fwhm_x"] = properties_df["x_fwhm"].std()
        key_values["resolution_theoretical_fwhm_lateral"] = self.get_metadata_values(
            "theoretical_fwhm_lateral_res"
        )
        key_values[
            "resolution_theoretical_fwhm_lateral_units"
        ] = self.get_metadata_units("theoretical_fwhm_lateral_res")
        key_values["resolution_theoretical_fwhm_axial"] = self.get_metadata_values(
            "theoretical_fwhm_axial_res"
        )
        key_values["resolution_theoretical_fwhm_axial_units"] = self.get_metadata_units(
            "theoretical_fwhm_axial_res"
        )

        self.output.append(
            model.KeyValues(
                name="Measurements_results",
                description="Output measurements",
                key_values=key_values,
            )
        )

        return True


# Calculate 2D FFT
# slice_2d = raw_img[17, ...].reshape([1, n_channels, x_size, y_size])
# fft_2D = fft_2d(slice_2d)

# Calculate 3D FFT
# fft_3D = fft_3d(spots_image)
#
# plt.imshow(np.log(fft_3D[2, :, :, 1]))  # , cmap='hot')
# # plt.imshow(np.log(fft_3D[2, 23, :, :]))  # , cmap='hot')
# plt.show()
#



"""This module takes care of the data model used by microscopemetrics.
It creates a few classes representing input data and output data
"""
from abc import ABC
from dataclasses import field
from pydantic.dataclasses import dataclass
from pydantic import validator, BaseConfig, create_model
from pydantic.color import Color

from pandas import DataFrame
from numpy import ndarray

from typing import Union, List, Tuple, Any


class MetadataConfig(BaseConfig):
    validate_assignment = True
    arbitrary_types_allowed = True


@dataclass
class MetricsDataset:
    """This class represents a single dataset including the intensity data and the name.
    Instances of this class are used by the analysis routines to get the necessary data to perform the analysis"""

    data: dict = field(default_factory=dict)
    metadata: dict = field(default_factory=dict)

    def add_data_requirement(
        self, name: str, description: str, data_type, optional: bool, replace: bool
    ):
        if not replace and name in self.data:
            raise KeyError(
                f"The key {name} is already used. Use argument replace=True to explicitly replace it"
            )

        model = create_model(
            name,
            value=(data_type, None),
            description=(str, description),
            optional=(bool, optional),
            __config__=MetadataConfig,
        )

        self.data[name] = model()
        setattr(self, name, self.data[name])

    def remove_data_requirement(self, name: str):
        try:
            del self.data[name]
        except KeyError as e:
            raise KeyError(f'Metadata "{name}" does not exist') from e

    def add_metadata_requirement(
        self,
        name: str,
        description: str,
        data_type,
        optional: bool,
        units: str = None,
        default: Any = None,
        replace: bool = False,
    ):
        if not replace and name in self.metadata:
            raise KeyError(
                f"The key {name} is already used. Use argument replace=True to explicitly replace it"
            )

        model = create_model(
            name,
            value=(data_type, default),
            description=(str, description),
            optional=(bool, optional),
            units=(str, units),
            default=(data_type, default),
            __config__=MetadataConfig,
        )

        self.metadata[name] = model()
        setattr(self, name, self.metadata[name])

    def remove_metadata_requirement(self, name: str):
        try:
            del self.metadata[name]
        except KeyError as e:
            raise KeyError(f'Metadata "{name}" does not exist') from e

    def describe_requirements(self):
        str_output = ["DATA requirements:\n"]
        for name, req in self.data.items():
            str_output.append("----------")
            str_output.append(f"Name: {name}")
            str_output.append(f"Description: {req.description}")
            str_output.append(f"Optional: {req.optional}")
        str_output.append("----------\n")
        str_output.append("METADATA requirements:\n")
        for name, req in self.metadata.items():
            str_output.append("----------")
            str_output.append("Name: " + name)
            str_output.append(f"Description: {req.description}")
            str_output.append(f"Optional: {req.optional}")
            str_output.append(f"Units: {req.units}")
            str_output.append(f"Default: {req.default}")
        str_output.append("----------")
        str_output = "\n".join(str_output)
        return str_output

    def list_unmet_requirements(self):
        unmet_data_req = [name for name, req in self.data.items() if req.value is None]
        unmet_metadata_req = [
            name
            for name, req in self.metadata.items()
            if not req.optional and req.value is None
        ]
        return unmet_data_req + unmet_metadata_req

    def validate_requirements(self):
        data_req_met = all(req.value is not None for _, req in self.data.items())
        metadata_req_met = all(
            req.value is not None
            for _, req in self.metadata.items()
            if not req.optional
        )
        return data_req_met and metadata_req_met

    def get_data_values(self, name: Union[str, list]):
        if isinstance(name, str):
            try:
                return self.data[name].value
            except KeyError as e:
                raise KeyError(f'Datum "{name}" does not exist') from e
        elif isinstance(name, list):
            return {k: self.get_data_values(k) for k in name}
        else:
            raise TypeError("get_metadata_values requires a string or list of strings")

    def get_metadata_values(self, name: Union[str, list]):
        if isinstance(name, str):
            try:
                return self.metadata[name].value
            except KeyError as e:
                raise KeyError(f'Metadatum "{name}" does not exist') from e
        elif isinstance(name, list):
            return {k: self.get_metadata_values(k) for k in name}
        else:
            raise TypeError("get_metadata_values requires a string or list of strings")

    def get_metadata_units(self, name: Union[str, list]):
        if isinstance(name, str):
            try:
                return self.metadata[name].units
            except KeyError as e:
                raise KeyError(f'Metadatum "{name}" does not exist') from e
        elif isinstance(name, list):
            return {k: self.get_metadata_units(k) for k in name}
        else:
            raise TypeError("get_metadata_units requires a string or list of strings")

    def get_metadata_defaults(self, name: Union[str, list]):
        if isinstance(name, str):
            try:
                return self.metadata[name].default
            except KeyError as e:
                raise KeyError(f'Metadatum "{name}" does not exist') from e
        elif isinstance(name, list):
            return {k: self.get_metadata_defaults(k) for k in name}
        else:
            raise TypeError("get_metadata_units requires a string or list of strings")

    def set_data_values(self, name: str, value):
        try:
            self.data[name].value = value
        except KeyError as e:
            raise KeyError(f'Metadata "{name}" is not a valid requirement') from e

    def del_data_values(self, name: str):
        try:
            self.data[name].value = None
        except KeyError as e:
            raise KeyError(f'Metadata "{name}" does not exist') from e

    def set_metadata_values(self, name: str, value):
        try:
            self.metadata[name].value = value
        except KeyError as e:
            raise KeyError(f'Metadata "{name}" is not a valid requirement') from e

    def del_metadata_values(self, name: str):
        try:
            self.metadata[name].value = None
        except KeyError as e:
            raise KeyError(f'Metadata "{name}" does not exist') from e

    def list_data_names(self):
        return [k for k, _ in self.data]

    def list_metadata_names(self):
        return [k for k, _ in self.metadata]


@dataclass
class MetricsOutput:
    """This class is used by microscopemetrics to return the output of an analysis."""

    description: str = field(default=None)
    properties: dict = field(default_factory=dict, init=False)

    def describe_properties(self):
        description = ""
        for n, p in self.properties.items():
            description += f"{n}\n"
            description += p.describe()
            description += "\n========\n"

        return description

    def get_property(self, name: str):
        return self.properties[name]

    def delete(self, name: str):
        del self.properties[name]

    def append(self, output_property):
        if isinstance(output_property, OutputProperty):
            self.properties.update({output_property.name: output_property})
        else:
            raise TypeError("Property appended must be a subtype of OutputProperty")

    def extend(self, properties_list: list):
        for p in properties_list:
            self.append(p)

    def _get_properties_by_type(self, property_type):
        return [v for _, v in self.properties.items() if v.type == property_type]

    def get_images(self):
        return self._get_properties_by_type("Image")

    def get_rois(self):
        return self._get_properties_by_type("Roi")

    def get_tags(self):
        return self._get_properties_by_type("Tag")

    def get_key_values(self):
        return self._get_properties_by_type("KeyValues")

    def get_tables(self):
        return self._get_properties_by_type("Table")

    def get_comments(self):
        return self._get_properties_by_type("Comment")


@dataclass
class OutputProperty(ABC):
    name: str
    description: str

    def describe(self):
        """
        Returns a pretty string describing the property. Includes name, type and description.
        :return: str
        """
        return (
            f"Name: {self.name}\n"
            f"Type: {self.type}\n"
            f"Description: {self.description}"
        )

    @property
    def type(self):
        """
        Returns the type of the OutputProperty as a string.
        :return: str
        """
        return self.__class__.__name__


@dataclass
class Image(OutputProperty):
    data: Any

    @validator("data", allow_reuse=True)
    def _is_ndarray(cls, d):
        if isinstance(d, ndarray):
            return d
        else:
            raise TypeError("image output must be a Numpy ndarray")


@dataclass
class Roi(OutputProperty):
    shapes: list


@dataclass
class Shape(ABC):
    z: float = field(default=None)
    c: int = field(default=None)
    t: int = field(default=None)
    fill_color: Color = field(default=Color((10, 10, 10, 0.1)))
    stroke_color: Color = field(default=Color((255, 255, 255, 1.0)))
    stroke_width: int = field(default=1)
    label: str = field(default=None)


@dataclass
class Point(Shape):
    x: float = field(default=None, metadata={"units": "PIXELS"})
    y: float = field(default=None, metadata={"units": "PIXELS"})


@dataclass
class Line(Shape):
    x1: float = field(default=None, metadata={"units": "PIXELS"})
    y1: float = field(default=None, metadata={"units": "PIXELS"})
    x2: float = field(default=None, metadata={"units": "PIXELS"})
    y2: float = field(default=None, metadata={"units": "PIXELS"})


@dataclass
class Rectangle(Shape):
    x: float = field(default=None, metadata={"units": "PIXELS"})
    y: float = field(default=None, metadata={"units": "PIXELS"})
    w: float = field(default=None, metadata={"units": "PIXELS"})
    h: float = field(default=None, metadata={"units": "PIXELS"})


@dataclass
class Ellipse(Shape):
    x: float = field(default=None, metadata={"units": "PIXELS"})
    y: float = field(default=None, metadata={"units": "PIXELS"})
    x_rad: float = field(default=None, metadata={"units": "PIXELS"})
    y_rad: float = field(default=None, metadata={"units": "PIXELS"})


@dataclass
class Polygon(Shape):
    points: List[Tuple[float, float]] = field(
        default=None, metadata={"units": "PIXELS"}
    )
    is_open: bool = field(default=False)


class Mask(Shape):
    pass


@dataclass
class Tag(OutputProperty):
    tag_value: str


@dataclass
class KeyValues(OutputProperty):
    key_values: dict

    @validator("key_values", allow_reuse=True)
    def _may_be_casted_to_str(cls, k_v):
        if all(isinstance(v, (str, int, float, list, tuple)) for _, v in k_v.items()):
            return k_v
        else:
            raise TypeError(
                "Values for a KeyValue property must be str, int, float, list or tuple"
            )


@dataclass
class Table(OutputProperty):
    table: Any

    @validator("table", allow_reuse=True)
    def _may_be_casted_to_df(cls, t):
        if isinstance(t, (DataFrame, dict)):
            return t
        else:
            raise TypeError("table must be a pandas DataFrame or a dict")


@dataclass
class Comment(OutputProperty):
    comment: str



# Main samples module defining the sample superclass

from abc import ABC, abstractmethod
import logging
from typing import Union, Any

from ..model import model

# We are defining some global dictionaries to register the different analysis types
IMAGE_ANALYSIS_REGISTRY = {}
DATASET_ANALYSIS_REGISTRY = {}
PROGRESSION_ANALYSIS_REGISTRY = {}


# Decorators to register exposed analysis functions
def register_image_analysis(cls):
    IMAGE_ANALYSIS_REGISTRY[cls.__name__] = cls
    return cls


def register_dataset_analysis(cls):
    DATASET_ANALYSIS_REGISTRY[cls.__name__] = cls
    return cls


def register_progression_analysis(cls):
    PROGRESSION_ANALYSIS_REGISTRY[cls.__name__] = cls
    return cls


# Create a logging service
logger = logging.getLogger(__name__)


class Configurator(ABC):
    """This is a superclass taking care of the configuration of a new sample. Helps generating configuration files and
    defines the metadata required for the different analyses. You should subclass this when you create a
    new sample. One for each type of configurator that you wish to have.
    """

    # The configuration section has to be defined for every subclass
    CONFIG_SECTION: str = None

    def __init__(self, config):
        self.config = config
        self.metadata_definitions = self.define_metadata()

    @abstractmethod
    def define_metadata(self):
        pass

    @classmethod
    def register_sample_analysis(cls, sample_class):
        cls.SAMPLE_CLASS = sample_class
        return sample_class


class Analysis(ABC):
    """This is the superclass defining the interface to a sample object. You should subclass this when you create a
    new sample analysis."""

    def __init__(self, output_description):
        self.input = model.MetricsDataset()
        self.output = model.MetricsOutput(description=output_description)

    @classmethod
    def get_name(cls):
        """Returns the module name of the class. Without path and extension.
        :returns a string with the module name
        """
        return cls.__module__.split(sep=".")[-1]

    def add_data_requirement(
        self,
        name: str,
        description: str,
        data_type,
        optional: bool = False,
        replace: bool = False,
    ):
        self.input.add_data_requirement(
            name=name,
            description=description,
            data_type=data_type,
            optional=optional,
            replace=replace,
        )

    def add_metadata_requirement(
        self,
        name: str,
        description: str,
        data_type,
        optional: bool,
        units: str = None,
        default: Any = None,
    ):
        self.input.add_metadata_requirement(
            name=name,
            description=description,
            data_type=data_type,
            optional=optional,
            units=units,
            default=default,
        )

    def describe_requirements(self):
        print(self.input.describe_requirements())

    def validate_requirements(self):
        return self.input.validate_requirements()

    def list_unmet_requirements(self):
        return self.input.list_unmet_requirements()

    def set_data(self, name: str, value):
        self.input.set_data_values(name, value)

    def set_metadata(self, name: str, value):
        self.input.set_metadata_values(name, value)

    def delete_data(self, name: str):
        self.input.del_data_values(name)

    def delete_metadata(self, name: str):
        self.input.del_metadata_values(name)

    def get_data_values(self, name: Union[str, list]):
        return self.input.get_data_values(name)

    def get_metadata_values(self, name: Union[str, list]):
        return self.input.get_metadata_values(name)

    def get_metadata_units(self, name: Union[str, list]):
        return self.input.get_metadata_units(name)

    def get_metadata_defaults(self, name: Union[str, list]):
        return self.input.get_metadata_defaults(name)

    @abstractmethod
    def run(self):
        raise NotImplemented()


class Reporter(ABC):
    """This is the superclass taking care of creating reports for a particular type of sample.
    You should subclass this when you create a new sample."""

    def __init__(
        self,
        config,
        image_report_to_func={},
        dataset_report_to_func={},
        microscope_report_to_func={},
    ):
        """Add to the init subclass a dictionary mapping analyses strings to functions
        :type config: analysis_config section
        :param config: analysis_config section specifying sample options
        :type image_report_to_func: dict
        :param image_report_to_func: dictionary mapping image analyses strings to functions
        :type dataset_report_to_func: dict
        :param dataset_report_to_func: dictionary mapping dataset analyses strings to functions
        :type microscope_report_to_func: dict
        :param microscope_report_to_func: dictionary mapping microscope analyses strings to functions
        """
        self.config = config
        self.image_report_to_func = image_report_to_func
        self.dataset_report_to_func = dataset_report_to_func
        self.microscope_report_to_func = microscope_report_to_func

    def produce_image_report(self, image):
        pass

    def produce_dataset_report(self, dataset):
        pass

    def produce_device_report(self, device):
        pass

    # TODO: move this where it belongs
    # # Helper functions
    # def get_tables(self, omero_object, namespace_start='', name_filter=''):
    #     tables_list = list()
    #     resources = omero_object._conn.getSharedResources()
    #     for ann in omero_object.listAnnotations():
    #         if isinstance(ann, gw.FileAnnotationWrapper) and \
    #                 ann.getNs().startswith(namespace_start) and \
    #                 name_filter in ann.getFileName():
    #             table_file = omero_object._conn.getObject("OriginalFile", attributes={'name': ann.getFileName()})
    #             table = resources.openTable(table_file._obj)
    #             tables_list.append(table)
    #
    #     return tables_list
    #



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['microscopemetrics',
 'microscopemetrics.analysis',
 'microscopemetrics.devices',
 'microscopemetrics.model',
 'microscopemetrics.samples',
 'microscopemetrics.utilities']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.1.5,<2.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'scikit-image>=0.19.3,<0.20.0',
 'scipy>=1.5.4,<2.0.0']

setup_kwargs = {
    'name': 'microscopemetrics',
    'version': '0.1.0',
    'description': 'A package providing analysis routines to measure the performance of micrsocopes used in biomedical research',
    'long_description': '# OMERO.metrics\nA python library to run micrsocope metrology\n\nTry it here [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/juliomateoslangerak/microscope-metrics/HEAD)\n',
    'author': 'Julio Mateos Langerak',
    'author_email': 'julio.matoes-langerak@igh.cnrs.fr',
    'maintainer': 'Julio Mateos Langerak',
    'maintainer_email': 'julio.matoes-langerak@igh.cnrs.fr',
    'url': 'https://github.com/MontpellierRessourcesImagerie/microscope-metrics',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""These are some possibly useful code snippets"""
from pandas import DataFrame
from skimage.filters import threshold_otsu, apply_hysteresis_threshold, gaussian
from skimage.segmentation import clear_border
from skimage.measure import label, regionprops
from skimage.morphology import closing, square, cube, octahedron, ball
from skimage.feature import peak_local_max
from scipy.spatial.distance import cdist
from scipy import ndimage
import numpy as np
from itertools import permutations

import logging

# Creating logging services
module_logger = logging.getLogger(__name__)


def _segment_channel(
    channel,
    min_distance,
    method,
    threshold,
    sigma,
    low_corr_factor,
    high_corr_factor,
    indices,
):
    """Segment a channel (3D numpy array)"""
    if threshold is None:
        threshold = threshold_otsu(channel)

    if sigma is not None:
        channel = gaussian(
            image=channel, multichannel=False, sigma=sigma, preserve_range=True
        )

    if method == "hysteresis":  # We may try here hysteresis threshold
        thresholded = apply_hysteresis_threshold(
            channel, low=threshold * low_corr_factor, high=threshold * high_corr_factor
        )

    elif method == "local_max":  # We are applying a local maxima algorithm
        peaks = peak_local_max(
            channel,
            min_distance=min_distance,
            threshold_abs=(threshold * 0.5),
            indices=indices,
        )
        thresholded = np.copy(channel)
        thresholded[peaks] = thresholded.max()
        thresholded = apply_hysteresis_threshold(
            thresholded,
            low=threshold * low_corr_factor,
            high=threshold * high_corr_factor,
        )
    else:
        raise Exception("A valid segmentation method was not provided")

    closed = closing(thresholded, cube(min_distance))
    cleared = clear_border(closed)
    return label(cleared)


def segment_image(
    image,
    min_distance,
    sigma=None,
    method="local_max",
    low_corr_factors=None,
    high_corr_factors=None,
    indices=False,
):
    """Segment an image and return a labels object.
    Image must be provided as zctxy numpy array
    """
    module_logger.info("Image being segmented...")

    if low_corr_factors is None:
        low_corr_factors = [0.95] * image.shape[1]
        module_logger.warning("No low correction factor specified. Using defaults")
    if high_corr_factors is None:
        high_corr_factors = [1.05] * image.shape[1]
        module_logger.warning("No high correction factor specified. Using defaults")

    if (
        len(high_corr_factors) != image.shape[1]
        or len(low_corr_factors) != image.shape[1]
    ):
        raise Exception(
            "The number of correction factors does not match the number of channels."
        )

    # We create an empty array to store the output
    labels_image = np.zeros(image.shape, dtype=np.uint16)
    for c in range(image.shape[1]):
        for t in range(image.shape[2]):
            labels_image[:, c, t, ...] = _segment_channel(
                image[:, c, t, ...],
                min_distance=min_distance,
                method=method,
                threshold=None,
                sigma=sigma,
                low_corr_factor=low_corr_factors[c],
                high_corr_factor=high_corr_factors[c],
                indices=indices,
            )
    return labels_image


def _compute_channel_spots_properties(
    channel, label_channel, remove_center_cross=False, pixel_size=None
):
    """Analyzes and extracts the properties of a single channel"""

    regions = regionprops(label_channel, channel)

    ch_properties = [
        {
            "label": region.label,
            "area": region.area,
            "centroid": region.centroid,
            "weighted_centroid": region.weighted_centroid,
            "max_intensity": region.max_intensity,
            "mean_intensity": region.mean_intensity,
            "min_intensity": region.min_intensity,
            "integrated_intensity": region.mean_intensity * region.area,
        }
        for region in regions
    ]

    if (
        remove_center_cross
    ):  # Argolight spots pattern contains a central cross that we might want to remove
        largest_area = 0
        largest_region = None
        for region in ch_properties:
            if region["area"] > largest_area:  # We assume the cross is the largest area
                largest_area = region["area"]
                largest_region = region
        if largest_region:
            ch_properties.remove(largest_region)
    ch_positions = np.array([x["weighted_centroid"] for x in ch_properties])
    if pixel_size:
        ch_positions = ch_positions[0:] * pixel_size

    return ch_properties, ch_positions


def compute_spots_properties(image, labels, remove_center_cross=False, pixel_size=None):
    """Computes a number of properties for the PSF-like spots found on an image provided they are segmented"""
    properties = []
    positions = []

    for c in range(image.shape[1]):
        for t in range(image.shape[2]):
            pr, pos = _compute_channel_spots_properties(
                channel=image[:, c, t, ...],
                label_channel=labels[:, c, t, ...],
                remove_center_cross=remove_center_cross,
                pixel_size=pixel_size,
            )
            properties.append(pr)
            positions.append(pos)

    return properties, positions


def compute_distances_matrix(positions, max_distance, pixel_size=None):
    """Calculates Mutual Closest Neighbour distances between all channels and returns the values as"""
    module_logger.info("Computing distances between spots")

    if len(positions) < 2:
        raise Exception("Not enough dimensions to do a distance measurement")

    channel_permutations = list(permutations(range(len(positions)), 2))

    if not pixel_size:  # TODO: make sure the units are corrected if no pixel size
        pixel_size = np.array((1, 1, 1))
        module_logger.warning("No pixel size specified. Using the unit")
    else:
        pixel_size = np.array(pixel_size)

    for a, b in channel_permutations:
        distances_matrix = cdist(positions[a], positions[b], w=pixel_size)

        distances_df = DataFrame()

        for i, (pos_A, d) in enumerate(zip(positions[a], distances_matrix)):
            if d.min() < max_distance:
                distances_df = distances_df.append(
                    {
                        "channel_a": a,
                        "channel_b": b,
                        "z_coord_a": pos_A[0],
                        "y_coord_a": pos_A[1],
                        "x_coord_a": pos_A[2],
                        "z_coord_b": positions[b][d.argmin()][0],
                        "y_coord_b": positions[b][d.argmin()][1],
                        "x_coord_b": positions[b][d.argmin()][2],
                        "z_dist": pos_A[0] - positions[b][d.argmin()][0],
                        "y_dist": pos_A[1] - positions[b][d.argmin()][1],
                        "x_dist": pos_A[2] - positions[b][d.argmin()][2],
                        "dist_3d": d.min(),
                        "labels_a": i,
                        "labels_b": d.argmin(),
                    },
                    ignore_index=True,
                )

    return distances_df


def _radial_mean(image, bins=None):
    """Computes the radial mean from an input 2d image.
    Taken from scipy-lecture-notes 2.6.8.4
    """
    # TODO: workout a binning = image size
    if not bins:
        bins = 200
    size_x, size_y = image.shape
    x, y = np.ogrid[0:size_x, 0:size_y]

    r = np.hypot(x - size_x / 2, y - size_y / 2)

    rbin = (bins * r / r.max()).astype(np.int)

    return ndimage.mean(image, labels=rbin, index=np.arange(1, rbin.max() + 1))


def _channel_fft_2d(channel):
    channel_fft = np.fft.rfft2(channel)
    return np.fft.fftshift(np.abs(channel_fft), axes=1)


def fft_2d(image):
    # Create an empty array to contain the transform
    fft = np.zeros(
        shape=(image.shape[1], image.shape[2], image.shape[3] // 2 + 1), dtype="float64"
    )
    for c in range(image.shape[2]):
        fft[c, :, :] = _channel_fft_2d(image[..., c, :, :])

    return fft


def _channel_fft_3d(channel):
    channel_fft = np.fft.rfftn(channel)
    return np.fft.fftshift(np.abs(channel_fft), axes=1)


def fft_3d(image):
    fft = np.zeros(
        shape=(
            image.shape[0],
            image.shape[1],
            image.shape[2],
            image.shape[3],
            image.shape[4] // 2 + 1,
        ),  # We only compute the real part of the FFT
        dtype="float64",
    )
    for c in range(image.shape[-3]):
        fft[..., c, :, :] = _channel_fft_3d(image[..., c, :, :])

    return fft
